from typing import List

from omegaconf.omegaconf import DictConfig, ListConfig

from .cache import load_cache as load_cache
from .cache import save_cache as save_cache
from .common import PATH as PATH
from .common import ConfigurattError, pop_conf
from .core import load as load
from .core import load_nested as load_nested

# * as * used to re-export symbols and satisfy linter.
from .deps import ConfigDependencies as ConfigDependencies


def check_requirements(conf: DictConfig, bases: List[DictConfig], strict: bool = True):
    # build requirements map first using recursive helper
    def _check_item(section, name):
        """Helper function, checks if a dotted name x.y.z is in conf, or .x is in section"""
        name_elems = name.split(".")
        # if name starts with ".", look in current section, else check config and any bases
        if name_elems[0] == "":
            name_elems = name_elems[1:]
            configs = [section]
        else:
            configs = [conf] + bases
        # try all configs
        for config in configs:
            # start looking at this config's top level
            section = config
            for elem in name_elems:
                # break loop if not found at this level
                if elem not in section:
                    break
                try:
                    section = section[elem]
                except Exception:
                    break
            # got to the end? Success
            else:
                return True
        # nothing found -- return False
        return False

    def _scan(section, location=""):
        # Short-circuit out if section is empty. Seems both DictConfig and ListConfig can be a kind of None (funny
        # OmegeConf API feature), where they're an instance of DictConfig or ListConfig, but don't support __getitem__.
        if not section:
            return [], False

        delete_self = False
        to_delete = []
        unresolved = []
        if isinstance(section, DictConfig):
            # get requirements from section
            reqs = pop_conf(section, "_requires", [])
            contingents = pop_conf(section, "_contingent", [])
            if type(reqs) is str:
                reqs = [reqs]
            if type(contingents) is str:
                contingents = [contingents]
            if not isinstance(reqs, (list, ListConfig)):
                raise ConfigurattError(f"'{location or ''}._requires' must be a string or a sequence")
            if not isinstance(contingents, (list, ListConfig)):
                raise ConfigurattError(f"'{location or ''}._contingent' must be a string or a sequence")
            # check requirements and contingencies
            for req in reqs:
                if not _check_item(section, req):
                    unresolved.append(
                        (location, req, ConfigurattError(f"section '{location}' has missing requirement '{req}'"))
                    )
            for cont in contingents:
                if not _check_item(section, cont):
                    unresolved.append((location, cont, None))
                    delete_self = True
            # recurse into content
            for name, value in section.items_ex(resolve=False):
                unres, delete = _scan(value, location=f"{location}.{name}" if location else name)
                unresolved += unres
                if delete:
                    to_delete.append(name)
            # delete what needs to be deleted
            for name in to_delete:
                del section[name]
        # lists -- recurse into content
        elif isinstance(section, ListConfig):
            for i, value in enumerate(section._iter_ex(resolve=False)):
                unres, delete = _scan(value, location=f"{location or ''}[{i}]")
                unresolved += unres
                if delete:
                    to_delete.append(i)
            # delete what needs to be deleted
            for i in to_delete[::-1]:
                del section[i]
            delete_self = False
        return unresolved, delete_self

    # build list of unresolved items
    unresolved, delete = _scan(conf)

    # top-level config wants to be deleted? this is an error, it's not supposed to be contingent
    if delete:
        raise ConfigurattError("top-level configuration is contingent on missing sections")

    # in strict mode, throw any unresolved errors
    if strict:
        errors = [exc for (_, _, exc) in unresolved if exc is not None]
        if errors:
            raise ConfigurattError("configuration has missing requirements", nested=errors)

    return unresolved



import importlib.metadata
import logging
import os
import subprocess

from . import exceptions
from .logging_utils import MultiplexingHandler

__version__ = importlib.metadata.version(__package__)


def init_logger(name="SCABHA", fmt="{asctime}: {message}", datefmt="%Y-%m-%d %H:%M:%S", loglevel="INFO"):
    """Returns the global Stimela logger (initializing if not already done so, with the given values)"""
    global log
    if log is None:
        log = logging.getLogger(name)
        log.propagate = False

        level = os.environ.get("SCABHA_LOG_LEVEL") or "INFO"
        log.setLevel(getattr(logging, level, logging.INFO))

        global log_console_handler, log_formatter

        log_formatter = logging.Formatter(fmt, datefmt, style="{")

        log_console_handler = MultiplexingHandler()
        log_console_handler.setFormatter(log_formatter)
        log_console_handler.setLevel(getattr(logging, loglevel))
        log.addHandler(log_console_handler)

        exceptions.set_logger(log)
    return log


def set_logger(logger):
    global log
    log = logger
    exceptions.set_logger(logger)


def logger():
    return init_logger()


def report_memory():
    """Reports memory status"""
    try:
        output = subprocess.check_output(["/usr/bin/free", "-h"]).decode().splitlines(keepends=False)
    except subprocess.CalledProcessError as exc:
        log.warning(f"/usr/bin/free -h exited with code {exc.returncode}")
        return
    for line in output:
        log.info(line)


def init():
    """Initiatlizes scabha standalone inside container"""
    log.info("Initial memory state:")
    report_memory()


log = None
init_logger()



import dataclasses
import importlib
import inspect
import re
import traceback

# Imports of typing and scabha.basetypes needed for calling eval on arbitrary type strings.
import typing
import warnings
from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Dict, List, Optional

import rich.markup
from omegaconf import DictConfig, ListConfig, OmegaConf
from rich.markdown import Markdown
from rich.table import Table

from scabha import basetypes
from scabha.basetypes import (
    UNSET,
    EmptyClassDefault,
    EmptyDictDefault,
    EmptyListDefault,
    is_file_list_type,
    is_file_type,
)

from .exceptions import (
    AssignmentError,
    DefinitionError,
    ParameterValidationError,
    SchemaError,
    StimelaPendingDeprecationWarning,
)
from .substitutions import SubstitutionNS
from .validate import Unresolved, validate_parameters

## almost supported by omegaconf, see https://github.com/omry/omegaconf/issues/144, for now just use Any
ListOrString = Any

Conditional = Optional[str]

# this marks unset defaults: it's turned into an UNSET by the __post_init__
# method (UNSET can't be used directly in an OmegeConf structured schema, hence
# the need for this stopgap)
_UNSET_DEFAULT = "<UNSET DEFAULT VALUE>"

warnings.simplefilter("default", category=StimelaPendingDeprecationWarning)

_DTYPE_EVAL_SAFESPACE = {}
_basetypes_typing_namespace = vars(basetypes) | vars(typing)
for _key, _val in _basetypes_typing_namespace.items():
    if inspect.isfunction(_val) or inspect.ismethod(_val) or _key.startswith("__"):
        continue
    elif isinstance(_val, type):
        _DTYPE_EVAL_SAFESPACE[_key] = _val
    elif getattr(_val, "__module__", False) in ["typing", "scabha.basetypes"]:
        _DTYPE_EVAL_SAFESPACE[_key] = _val
    # this condition has to come after looking for __module__


@dataclass
class ParameterPolicies(object):
    """This class describes policies that determine how a Parameter is turned into
    cab arguments. Most policies refer to how command-line arguments are formed up,
    although some also apply to Python callable cabs.
    """

    # if true, parameter is passed as key=value, not command line option
    key_value: Optional[bool] = None
    # if true, value is passed as a positional argument, not an option
    positional: Optional[bool] = None
    # if true, value is head-positional, i.e. passed *before* any options
    positional_head: Optional[bool] = None
    # for list-type values, use this as a separator to paste them together into one argument. Otherwise:
    #  * use "list" to pass list-type values as multiple arguments (--option X Y)
    #  * use "[]" to pass list-type values as a list  (--option [X,Y])
    #  * use "repeat" to repeat the option (--option X --option Y)
    repeat: Optional[str] = None
    # prefix for non-positional arguments
    prefix: Optional[str] = None

    # skip: does the parameter need to be passed as an argument to the underlying cargo
    # If skip is left as None, apply default logic, namely:
    # * inputs and named file-type outputs are not skipped (i.e. passed)
    # * all other outputs are skipped (i.e. not passed)
    # Set this to False or True to enforce (not) skipping rather
    skip: Optional[bool] = None
    # Same thing, but overrides the skip setting for implicit input and outputs
    # (if skip is not None, it takes priority, otherwise if I/O is implicit, this setting is applied)
    skip_implicits: Optional[bool] = None

    # if set, {}-substitutions on this paramater will not be done
    disable_substitutions: Optional[bool] = None

    # how to pass boolean True values. None = pass option name alone, else pass option name + given value
    explicit_true: Optional[str] = None
    # how to pass boolean False values. None = skip option, else pass option name + given value
    explicit_false: Optional[str] = None
    # pass flag value. None|False = skip, --<flag> true/false | 1/0 | yes/no | t/f
    explicit_flag: Optional[bool] = None
    # flag without the '--no-<flag>' option. None|False = skip
    is_flag: Optional[bool] = None

    # if set, a string-type value will be split into a list of arguments using this separator
    split: Optional[str] = None

    # dict of character replacements for mapping parameter name to command line
    replace: Optional[Dict[str, str]] = None

    # Value formatting policies.
    # If set, specifies {}-type format strings used to convert the value(s) to string(s).
    # For a non-list value:
    #   * if 'format_list_scalar' is set, formats the value into a list of strings as fmt[i].format(value, **dict)
    #     example:  ["{0}", "{0}"] will simply repeat the value twice
    #   * if 'format' is set, value is formatted as format.format(value, **dict)
    # For a list-type value:
    #   * if 'format_list' is set, each element #i formatted separately as fmt[i].format(*value, **dict)
    #     example:  ["{0}", "{2}"] will output elements 0 and 2, and skip element 1
    #   * if 'format' is set, each element #i is formatted as format.format(value[i], **dict)
    # **dict contains all parameters passed to a cab, so these can be used in the formatting
    format: Optional[str] = None
    format_list: Optional[List[str]] = None
    format_list_scalar: Optional[List[str]] = None

    # for Python callable cabs: if set, then missing parameters are passed as None values
    # if not set, missing parameters are not passed at all
    pass_missing_as_none: Optional[bool] = None


@dataclass
class PathPolicies(object):
    """This class describes policies for paths"""

    # if true, creates parent directories of output
    mkdir_parent: bool = True
    # if True, and parameter is a path, access to its parent directory is required
    access_parent: bool = False
    # if True, and parameter is a path, access to its parent directory is required in writable mode
    write_parent: bool = False
    # If True, and the output exists, remove before running
    remove_if_exists: bool = False


# This is used to classify parameters, for cosmetic and help purposes.
# Usually set automatically based on whether a parameter is required, whether a default is provided, etc.
ParameterCategory = IntEnum(
    "ParameterCategory", dict(Required=0, Optional=1, Implicit=2, Obscure=3, Hidden=4), module=__name__
)


@dataclass
class Parameter(object):
    """Parameter (of cab or recipe)"""

    info: str = ""
    # for input parameters, this flag indicates a read-write (aka input-output aka mixed-mode) parameter e.g. an MS
    writable: bool = False
    # data type
    dtype: str = "str"
    # specifies that the value is implicitly set inside the step (i.e. not a free parameter).
    # Typically used with filenames,
    implicit: Any = None
    # optonal list of arbitrary tags, used to group parameters
    tags: List[str] = EmptyListDefault()

    # If True, parameter is required. None/False, not required.
    # For outputs, required=False means missing output will not be treated as an error.
    # For aliases, False at recipe level will override the target setting, while the default of None won't.
    required: Optional[bool] = None

    # restrict value choices, i.e. making for an option-type parameter
    choices: Optional[List[Any]] = ()

    # for List or Dict-type parameters, restict values of list elements or dict entries to a list of choices
    element_choices: Optional[List[Any]] = None

    # default value
    default: Any = _UNSET_DEFAULT

    # list of aliases for this parameter (references to other parameters whose schemas/values this parameter shares)
    aliases: Optional[List[str]] = ()

    # if true, create empty directory for the output itself, if it doesn't exist
    # will probably be deprecated in favour of path_policies.mkdir in the future
    mkdir: bool = False
    # additional policies related to path-type inputs and outputs
    path_policies: PathPolicies = EmptyClassDefault(PathPolicies)

    # these are deprecated in favour of path_policies
    remove_if_exists: Optional[bool] = None
    access_parent_dir: Optional[bool] = None
    write_parent_dir: Optional[bool] = None

    # for file and dir-type parameters: if True, the file(s)/dir(s) must exist. If False, they can be missing.
    # if None, then the default logic applies: inputs must exist, and outputs don't
    # May be deprecated in favour of path_policies.must_exist in the future
    must_exist: Optional[bool] = None

    # for file and dir-type parameters: if True, ignore them when making processing logic decisions based
    # on file freshness
    skip_freshness_checks: Optional[bool] = None

    # if command-line option for underlying binary has a different name, specify it here
    nom_de_guerre: Optional[str] = None

    # policies object, specifying a non-default way to handle this parameter
    policies: ParameterPolicies = EmptyClassDefault(ParameterPolicies)

    # Parameter category, purely cosmetic, used for generating help and debug messages.
    # Assigned automatically if None, but a schema may explicitly mark parameters as e.g.
    # "obscure" or "hidden"
    category: Optional[ParameterCategory] = None

    # metavar corresponding to this parameter. Used when constructing command-line interfaces
    metavar: Optional[str] = None

    # abbreviated option name for this parameter.  Used when constructing command-line interfaces
    abbreviation: Optional[str] = None

    # arbitrary metadata associated with parameter
    metadata: Dict[str, Any] = EmptyDictDefault()

    # If True, when constructing a CLI from the schema, omit the default value (if any).
    # Useful when the tool constructs itw own default values.
    suppress_cli_default: bool = False

    def __post_init__(self):
        def natify(value):
            # convert OmegaConf lists and dicts to native types
            if type(value) in (list, ListConfig):
                return [natify(x) for x in value]
            elif type(value) in (dict, DictConfig):
                return dict([(name, natify(value)) for name, value in value.items()])
            elif value is _UNSET_DEFAULT:
                return UNSET
            return value

        self.default = natify(self.default)
        self.choices = natify(self.choices)

        # check for deprecated settings
        if self.remove_if_exists is not None:
            warnings.warn(  # deprecated parameter definition
                "the remove_if_exists parameter property will be deprecated "
                "in favour of path_policies.remove_if_exists in a future release",
                StimelaPendingDeprecationWarning,
                stacklevel=0,
            )
            self.path_policies.remove_if_exists = self.remove_if_exists
        if self.access_parent_dir is not None:
            warnings.warn(  # deprecated parameter definition
                "the access_parent_dir parameter property will be deprecated "
                "in favour of path_policies.access_parent in a future release",
                StimelaPendingDeprecationWarning,
                stacklevel=0,
            )
            self.path_policies.access_parent = self.access_parent_dir
        if self.write_parent_dir is not None:
            warnings.warn(  # deprecated parameter definition
                "the write_parent_dir parameter property will be deprecated "
                "in favour of path_policies.write_parent in a future release",
                StimelaPendingDeprecationWarning,
                stacklevel=0,
            )
            self.path_policies.write_parent = self.write_parent_dir

        try:
            self._dtype = self._type_eval()
        except Exception as exc:
            raise SchemaError(f"'{self.dtype}' is not a valid dtype", exc)

        self._is_file_type = is_file_type(self._dtype)
        self._is_file_list_type = is_file_list_type(self._dtype)

        self._is_input = True

    def _type_eval(self, type_string: str = None, maxlen=256):
        """
        Evaluate a string type hint safely. This uses the `eval` function,
        so we perform basic tests to avoid malicious strings
        """
        type_string = type_string or self.dtype
        # remove functions and other callables

        # check for hints of bad intent
        malicious_hints = "os. sys. import eval exec compile globals() locals()".split()
        for hint in malicious_hints:
            if hint in type_string:
                raise ValueError(f"Input type '{type_string}' contains a potentially malicious string: {hint} ")
        # Default set to 256 chars, which should be more than enough for a type hint
        if len(type_string) > maxlen:
            raise ValueError(
                f"The length of the type '{type_string}' exceeds the maximum length ({maxlen})"
                "set to avoid injection of malicious code"
            )

        return eval(type_string, _DTYPE_EVAL_SAFESPACE)

    def get_category(self):
        """Returns category of parameter, auto-setting it if not already preset"""
        if self.category is None:
            if self.required:
                self.category = ParameterCategory.Required
            elif self.implicit is not None:
                self.category = ParameterCategory.Implicit
            else:
                self.category = ParameterCategory.Optional
        return self.category

    @property
    def is_input(self):
        return self._is_input

    @property
    def is_output(self):
        return not self._is_input

    @property
    def is_file_type(self):
        """True if parameter is a file or directory type"""
        return self._is_file_type

    @property
    def is_file_list_type(self):
        """True if parameter is a file or directory type"""
        return self._is_file_list_type

    @property
    def is_named_output(self):
        """True if parameter is a named file or directory output"""
        return self.is_output and self.is_file_type and not self.implicit


ParameterSchema = OmegaConf.structured(Parameter)

ParameterFields = set(f.name for f in dataclasses.fields(Parameter))


@dataclass
class Cargo(object):
    name: Optional[str] = None  # cab name (if None, use image or command name)
    fqname: Optional[str] = None  # fully-qualified name (recipe_name.step_label.etc.etc.)

    info: Optional[str] = None  # help string

    extra_info: Dict[str, str] = EmptyDictDefault()  # optional, additional help sections

    # schemas are postentially nested (dicts of dicts), which omegaconf doesn't quite recognize,
    # (or in my ignorance I can't specify it -- in any case Union support is weak), so do a dict to Any
    # "Leaf" elements of the nested dict must be Parameters
    inputs: Dict[str, Any] = EmptyDictDefault()
    outputs: Dict[str, Any] = EmptyDictDefault()
    defaults: Dict[str, Any] = EmptyDictDefault()

    dynamic_schema: Optional[str] = None  # function to call to augment inputs/outputs dynamically

    @staticmethod
    def flatten_schemas(io_dest, io, label, prefix=""):
        for name, value in io.items():
            if name == "subsection":
                continue
            name = f"{prefix}{name}"
            if not isinstance(value, Parameter):
                if isinstance(value, str):
                    schema = {}
                    value = value.strip()
                    # if value ends with a double-quoted string, parse out the docstring
                    if value.endswith('"') and '"' in value[:-1]:
                        value, info, _ = value.rsplit('"', 2)
                        value = value.strip()
                        schema["info"] = info
                    # does value contain "="? Parse it as "type = default" then
                    if "=" in value:
                        value, default = value.split("=", 1)
                        value = value.strip()
                        default = default.strip()
                        if (default.startswith('"') and default.endswith('"')) or (
                            default.startswith("'") and default.endswith("'")
                        ):
                            default = default[1:-1]
                        schema["default"] = default
                    # does value end with "*"? Mark as required
                    elif value.endswith("*"):
                        schema["required"] = True
                        value = value[:-1]
                    schema["dtype"] = value
                    io_dest[name] = Parameter(**schema)
                # else proper dict schema, or subsection
                else:
                    if not isinstance(value, (DictConfig, dict)):
                        raise SchemaError(f"{label}.{name} is not a valid schema")
                    # try to treat as Parameter based on field names
                    if not (set(value.keys()) - ParameterFields):
                        try:
                            value = OmegaConf.unsafe_merge(ParameterSchema.copy(), value)
                            io_dest[name] = Parameter(**value)
                        except Exception as exc0:
                            raise SchemaError(f"{label}.{name} is not a valid parameter definition", exc0) from None
                    # else assume subsection and recurse in
                    else:
                        try:
                            Cargo.flatten_schemas(io_dest, value, label=label, prefix=f"{name}.")
                        except SchemaError as exc:
                            raise SchemaError(
                                f"{label}.{name} was interpreted as nested section, but contains errors", exc
                            ) from None
        return io_dest

    def flatten_param_dict(self, output_params, input_params, prefix=""):
        for name, value in input_params.items():
            name = f"{prefix}{name}"
            if isinstance(value, (dict, DictConfig)):
                # if prefix.name. is present in schemas, treat as nested mapping
                if any(k.startswith(f"{name}.") for k in self.inputs_outputs):
                    self.flatten_param_dict(output_params, value, prefix=f"{name}.")
                    continue
            output_params[name] = value
        return output_params

    def __post_init__(self):
        self.fqname = self.fqname or self.name
        # flatten inputs/outputs into a single dict (with entries like sub.foo.bar)
        self.inputs = Cargo.flatten_schemas({}, self.inputs, "inputs")
        self.outputs = Cargo.flatten_schemas({}, self.outputs, "outputs")
        for schema in self.outputs.values():
            schema._is_input = False
        for name in self.inputs.keys():
            if name in self.outputs:
                raise DefinitionError(f"parameter '{name}' appears in both inputs and outputs")
        self._inputs_outputs = None
        self._implicit_params = set()  # marks implicitly set values
        # flatten defaults and aliases
        self.defaults = self.flatten_param_dict({}, self.defaults)
        # pausterized name
        self.name_ = re.sub(r"\W", "_", self.name or "")  # pausterized name
        # config and logger objects
        self.config = self.log = self.logopts = None
        # resolve callable for dynamic schemas
        self._dyn_schema = None
        if self.dynamic_schema is not None:
            if "." not in self.dynamic_schema:
                raise DefinitionError(f"{self.dynamic_schema}: module_name.function_name expected")
            modulename, funcname = self.dynamic_schema.rsplit(".", 1)
            try:
                mod = importlib.import_module(modulename)
            except ImportError as exc:
                raise DefinitionError(f"can't import {modulename}: {exc}")
            self._dyn_schema = getattr(mod, funcname, None)
            if not callable(self._dyn_schema):
                raise DefinitionError(f"{modulename}.{funcname} is not a valid callable")
            # make backup copy of original inputs/outputs
            self._original_inputs_outputs = self.inputs.copy(), self.outputs.copy()

    @property
    def inputs_outputs(self):
        if self._inputs_outputs is None:
            self._inputs_outputs = self.inputs.copy()
            self._inputs_outputs.update(**self.outputs)
        return self._inputs_outputs

    @property
    def finalized(self):
        return self.config is not None

    def unresolved_params(self, params):
        """Returns list of unresolved parameters"""
        return [name for name, value in params.items() if isinstance(value, Unresolved)]

    def finalize(self, config=None, log=None, fqname=None, nesting=0):
        if not self.finalized:
            if fqname is not None:
                self.fqname = fqname
            self.config = config
            self.nesting = nesting
            self.log = log
            self.logopts = config.opts.log.copy()

    @property
    def has_dynamic_schemas(self):
        return bool(self._dyn_schema)

    def apply_dynamic_schemas(self, params, subst: Optional[SubstitutionNS] = None):
        # update schemas, if dynamic schema is enabled
        if self._dyn_schema:
            # delete implicit parameters, since they may have come from older version of schema
            params = self._delete_implicit_parameters(params, subst)
            # get rid of unsets
            params = {key: value for key, value in params.items() if value is not UNSET and type(value) is not UNSET}
            try:
                self.inputs, self.outputs = self._dyn_schema(params, *self._original_inputs_outputs)
            except Exception:
                lines = traceback.format_exc().strip().split("\n")
                raise SchemaError("error evaluating dynamic schema", lines)  # [exc, sys.exc_info()[2]])
            self._inputs_outputs = None  # to regenerate
            for io in self.inputs, self.outputs:
                for name, schema in list(io.items()):
                    if isinstance(schema, DictConfig):
                        try:
                            schema = OmegaConf.unsafe_merge(ParameterSchema.copy(), schema)
                        except Exception as exc:
                            raise SchemaError("error in dynamic schema for parameter 'name'", exc)
                        io[name] = Parameter(**schema)
            # new outputs may have been added
            for schema in self.outputs.values():
                schema._is_input = False
            # re-resolve implicits
            self._resolve_implicit_parameters(params, subst)

    def _delete_implicit_parameters(self, params, subst: Optional[SubstitutionNS] = None):
        current = subst and getattr(subst, "current", None)
        for p in self._implicit_params:
            if p in params:
                del params[p]
            if current and p in current:
                del current[p]
        self._implicit_params = set()
        return params

    def _resolve_implicit_parameters(self, params, subst: Optional[SubstitutionNS] = None):
        # remove previously defined implicits
        self._delete_implicit_parameters(params, subst)
        # regenerate
        current = subst and getattr(subst, "current", None)
        for name, schema in self.inputs_outputs.items():
            if schema.implicit is not None and type(schema.implicit) is not Unresolved:
                if name in params and name not in self._implicit_params and params[name] != schema.implicit:
                    raise ParameterValidationError(f"implicit parameter {name} was supplied explicitly")
                if name in self.defaults:
                    raise SchemaError(f"implicit parameter {name} also has a default value")
                params[name] = schema.implicit
                self._implicit_params.add(name)
                if current:
                    current[name] = schema.implicit

    def prevalidate(self, params: Optional[Dict[str, Any]], subst: Optional[SubstitutionNS] = None):
        """Does pre-validation.
        No parameter substitution is done, but will check for missing params and such.
        A dynamic schema, if defined, is applied at this point."""
        self.finalize()
        # add implicits, if resolved
        self._resolve_implicit_parameters(params, subst)
        # assign unset categories
        for schema in self.inputs_outputs.values():
            schema.get_category()

        params = validate_parameters(
            params,
            self.inputs_outputs,
            defaults=self.defaults,
            subst=subst,
            fqname=self.fqname,
            check_unknowns=True,
            check_required=False,
            check_inputs_exist=False,
            check_outputs_exist=False,
            create_dirs=False,
            ignore_subst_errors=True,
        )

        return params

    def validate_inputs(
        self, params: Dict[str, Any], subst: Optional[SubstitutionNS] = None, loosely=False, remote_fs=False
    ):
        """Validates inputs.
        If loosely is True, then doesn't check for required parameters, and doesn't check for files to exist etc.
        This is used when skipping a step.
        If remote_fs is True, doesn't check files and directories.
        """
        assert self.finalized
        self._resolve_implicit_parameters(params, subst)

        # check inputs
        params1 = validate_parameters(
            params,
            self.inputs,
            defaults=self.defaults,
            subst=subst,
            fqname=self.fqname,
            check_unknowns=False,
            check_required=not loosely,
            check_inputs_exist=not loosely and not remote_fs,
            check_outputs_exist=False,
            create_dirs=not loosely and not remote_fs,
        )
        # check outputs
        params1.update(
            **validate_parameters(
                params,
                self.outputs,
                defaults=self.defaults,
                subst=subst,
                fqname=self.fqname,
                check_unknowns=False,
                check_required=False,
                check_inputs_exist=not loosely and not remote_fs,
                check_outputs_exist=False,
                create_dirs=not loosely and not remote_fs,
            )
        )
        return params1

    def validate_outputs(
        self, params: Dict[str, Any], subst: Optional[SubstitutionNS] = None, loosely=False, remote_fs=False
    ):
        """Validates outputs. Parameter substitution is done.
        If loosely is True, then doesn't check for required parameters, and doesn't check for files to exist etc.
        If remote_fs is True, doesn't check files and directories.
        """
        assert self.finalized
        # update implicits that weren't marked as unresolved
        for name in self._implicit_params:
            impl = self.inputs_outputs[name].implicit
            if type(impl) is not Unresolved:
                params[name] = self.inputs_outputs[name].implicit
        params.update(
            **validate_parameters(
                params,
                self.outputs,
                defaults=self.defaults,
                subst=subst,
                fqname=self.fqname,
                check_unknowns=False,
                check_required=not loosely,
                check_inputs_exist=not loosely and not remote_fs,
                check_outputs_exist=not loosely and not remote_fs,
            )
        )
        return params

    def rich_help(self, tree, max_category=ParameterCategory.Optional):
        """Generates help into a rich.tree.Tree object"""
        if self.info:
            tree.add("Description:").add(Markdown(self.info))
        # extra documentation?
        for section, content in self.extra_info.items():
            if not section.lower().endswith("inputs") and not section.lower().endswith("outputs"):
                tree.add(f"{section}:").add(Markdown(content))
        # adds tables for inputs and outputs
        for io, title in (self.inputs, "inputs"), (self.outputs, "outputs"):
            # add extra help sections
            for section, content in self.extra_info.items():
                if section.lower().endswith(title):
                    tree.add(f"{section}:").add(Markdown(content))
            # add parameters by category
            for cat in ParameterCategory:
                schemas = [(name, schema) for name, schema in io.items() if schema.get_category() == cat]
                if not schemas:
                    continue
                if cat > max_category:
                    subtree = tree.add(f"[dim]{cat.name} {title}: omitting {len(schemas)}[/dim]")
                    continue
                subtree = tree.add(f"{cat.name} {title}:")
                table = Table.grid("", "", "", padding=(0, 2))
                subtree.add(table)
                for name, schema in schemas:
                    attrs = []
                    default = self.defaults.get(name, schema.default)
                    if schema.implicit:
                        attrs.append(f"implicit: {schema.implicit}")
                    if default is not UNSET and not isinstance(default, Unresolved):
                        attrs.append(f"default: {default}")
                    if schema.choices:
                        attrs.append(f"choices: {', '.join(schema.choices)}")
                    info = []
                    schema.info and info.append(rich.markup.escape(schema.info))
                    attrs and info.append(f"[dim]\\[{rich.markup.escape(', '.join(attrs))}][/dim]")
                    table.add_row(
                        f"[bold]{name}[/bold]", f"[dim]{rich.markup.escape(str(schema.dtype))}[/dim]", " ".join(info)
                    )

    def assign_value(self, key: str, value: Any):
        """assigns a parameter value to the cargo.
        Recipe will override this to handle nested assignments. Cabs can't be assigned to
        (it will be handled by the wraping step)
        """
        raise AssignmentError(f"{self.name}: invalid assignment {key}={value}")

    @staticmethod
    def add_parameter_summary(params: Dict[str, Any], lines: Optional[List[str]] = None):
        if lines is None:
            lines = []
        for name, value in params.items():
            if isinstance(value, (list, tuple)) and len(value) > 10:
                sep1, sep2 = "()" if isinstance(value, tuple) else "[]"
                lines.append(f"  {name} = {sep1}{value[0]}, {value[1]}, ..., {value[-1]}{sep2}")
            else:
                lines.append(f"  {name} = {value}")
        return lines



import dataclasses
import keyword
import os.path
import pathlib
import re
from typing import Any, Dict, List, Optional

import pydantic
import pydantic.dataclasses
import yaml
from omegaconf import DictConfig, ListConfig, OmegaConf

from scabha.basetypes import Unresolved

from .basetypes import MS, UNSET, URI, Directory, File
from .evaluator import Evaluator
from .exceptions import Error, ParameterValidationError, SubstitutionErrorList
from .substitutions import SubstitutionNS, substitutions_from


def join_quote(values):
    return "'" + "', '".join(values) + "'" if values else ""


def evaluate_and_substitute_object(obj: Any, subst: SubstitutionNS, recursion_level: int = 1, location: List[str] = []):
    with substitutions_from(subst, raise_errors=True) as context:
        evaltor = Evaluator(subst, context, location=location, allow_unresolved=False)
        return evaltor.evaluate_object(obj, raise_substitution_errors=True, recursion_level=recursion_level)


def evaluate_and_substitute(
    inputs: Dict[str, Any],
    subst: SubstitutionNS,
    corresponding_ns: SubstitutionNS,
    defaults: Dict[str, Any] = {},
    ignore_subst_errors: bool = False,
    location: List[str] = [],
):
    with substitutions_from(subst, raise_errors=True) as context:
        evaltor = Evaluator(subst, context, location=location, allow_unresolved=True)
        inputs = evaltor.evaluate_dict(
            inputs, corresponding_ns=corresponding_ns, defaults=defaults, raise_substitution_errors=False
        )
        # collect errors
        if not ignore_subst_errors:
            errors = []
            for value in inputs.values():
                if type(value) is Unresolved:
                    errors += value.errors
            # check for substitution errors
            if errors:
                raise SubstitutionErrorList("unresolved {}-substitutions", errors)
    return inputs


def validate_parameters(
    params: Dict[str, Any],
    schemas: Dict[str, Any],
    defaults: Optional[Dict[str, Any]] = None,
    subst: Optional[SubstitutionNS] = None,
    fqname: str = "",
    check_unknowns=True,
    check_required=True,
    check_inputs_exist=True,
    check_outputs_exist=True,
    create_dirs=False,
    ignore_subst_errors=False,
) -> Dict[str, Any]:
    """Validates a dict of parameter values against a given schema

    Args:
        params (Dict[str, Any]):   map of input parameter values
        schemas (Dict[str, Any]):   map of parameter names to schemas. Each schema must contain a dtype field and a
            choices field.
        defaults (Dict[str, Any], optional): dictionary of default values to be used when a value is missing
        subst (SubsititionNS, optional): namespace to do {}-substitutions on parameter values
        fqname: fully-qualified name of the parameter set (e.g. "recipe_name.step_name"), used in error messages. If
            not given, errors will report parameter names only

        check_unknowns (bool): if True, unknown parameters (not in schema) raise an error
        check_required (bool): if True, missing parameters with required=True will raise an error
        check_inputs_exist (bool): if True, input files with must_exist={None,True} in schema must exist, or will
            raise an error. If False, doesn't check.
        check_outputs_exist (bool): if True, output files with must_exist=True in schema must exist, or will raise an
            error. If False, doesn't check.
        create_dirs (bool): if True, non-existing directories in filenames (and parameters with mkdir=True in schema)
                            will be created.
        ignore_subst_errors (bool): if True, substitution errors will be ignored


    Raises:
        ParameterValidationError: parameter fails validation
        SchemaError: bad schema
        SubstitutionErrorList: list of substitution errors, if they occur

    Returns:
        Dict[str, Any]: validated dict of parameters

    TODO:
        add options to propagate all errors out (as values of type Error) in place of exceptions?
    """
    # define function for converting parameter name into "fully-qualified" name
    if fqname:
        mkname = lambda name: f"{fqname}.{name}"  # noqa: E731
    else:
        mkname = lambda name: name  # noqa: E731

    # check for unknowns
    if check_unknowns:
        for name in params:
            if name not in schemas:
                raise ParameterValidationError(f"unknown parameter '{mkname(name)}'")

    # only evaluate the subset for which we have schemas
    inputs = dict((name, value) for name, value in params.items() if name in schemas)

    # build dict of all defaults
    all_defaults = {
        name: schema.default
        for name, schema in schemas.items()
        if schema.default is not UNSET and schema.default != "UNSET"
    }
    if defaults:
        all_defaults.update(**{name: value for name, value in defaults.items() if name in schemas})

    # update missing inputs from defaults
    inputs.update(**{name: value for name, value in all_defaults.items() if name not in inputs})

    # update implicit values
    #

    # perform substitution
    if subst is not None:
        inputs = evaluate_and_substitute(
            inputs,
            subst,
            subst.current,
            defaults=all_defaults,
            ignore_subst_errors=ignore_subst_errors,
            location=[fqname],
        )

    # split inputs into unresolved substitutions, and proper inputs
    unresolved = {name: value for name, value in inputs.items() if isinstance(value, Unresolved)}
    inputs = {name: value for name, value in inputs.items() if not isinstance(value, Unresolved)}

    # check that required args are present
    if check_required:
        missing = [
            mkname(name)
            for name, schema in schemas.items()
            if schema.required and inputs.get(name) is UNSET and name not in unresolved
        ]
        if missing:
            raise ParameterValidationError(f"missing required parameters: {join_quote(missing)}")

    # create dataclass from parameter schema
    validated = {}
    fields = []

    # maps parameter names to/from field names. Fields have "_" not "-"
    name2field = {}
    field2name = {}

    for name, schema in schemas.items():
        value = inputs.get(name, UNSET)
        if value is not UNSET:
            # sanitize name: dataclass won't take hyphens or periods
            fldname = orig_fldname = re.sub("\\W", "_", name)
            # avoid Python keywords and clashes with other field names by adding _x as needed
            num = 0
            while (
                keyword.iskeyword(fldname)
                or (hasattr(keyword, "issoftkeyword") and keyword.issoftkeyword(fldname))
                or fldname in field2name
            ):
                fldname += f"{orig_fldname}_{num}"
                num += 1
            # add to mapping
            field2name[fldname] = name
            name2field[name] = fldname

            fields.append((fldname, schema._dtype))

            # OmegaConf dicts/lists need to be converted to standard containers for pydantic to take them
            if isinstance(value, (ListConfig, DictConfig)):
                inputs[name] = OmegaConf.to_container(value)

    dcls = dataclasses.make_dataclass("Parameters", fields)

    # convert this to a pydantic dataclass which does validation
    pcls = pydantic.dataclasses.dataclass(dcls)

    # check Files etc.
    for name, value in list(inputs.items()):
        # get schema from those that need validation, skip if not in schemas
        schema = schemas.get(name)
        if schema is None:
            continue
        # skip errors
        if value is UNSET or isinstance(value, Error):
            continue
        dtype = schema._dtype

        # must this file exist? Schema may force this check, otherwise follow the default check_exist policy
        if schema.is_input:
            must_exist = check_inputs_exist and schema.must_exist is not False
        elif schema.is_output:
            must_exist = check_outputs_exist and schema.must_exist

        if schema.is_file_type or schema.is_file_list_type:
            # match to existing file(s)
            if type(value) is str:
                # try to interpret string as a formatted list (a list substituted in would come out like that)
                try:
                    files = yaml.safe_load(value)
                    if type(files) is not list:
                        files = [value]
                except Exception:
                    files = [value]
            elif isinstance(value, (list, tuple)):
                files = value
            else:
                raise ParameterValidationError(f"'{mkname(name)}={value}': invalid type '{type(value)}'")
            # convert to appropriate type
            files = [URI(f) for f in files]

            # check for existence of all files in list, if needed
            if must_exist:
                if not files:
                    raise ParameterValidationError(f"'{mkname(name)}': file(s) don't exist")
                not_exists = [uri.path for uri in files if not uri.remote and not os.path.exists(uri.path)]
                if not_exists:
                    raise ParameterValidationError(f"'{mkname(name)}': {','.join(not_exists)} doesn't exist")

            # check for single file/dir
            if schema.is_file_type:
                if len(files) > 1:
                    raise ParameterValidationError(f"'{mkname(name)}': multiple files given ({value})")
                # no files? must_exist was checked above, so return empty filename
                elif not files:
                    inputs[name] = ""
                # else one file/dir as expected, check it
                else:
                    # check that files are files and dirs are dirs
                    uri = files[0]
                    if not uri.remote and os.path.exists(uri.path):
                        if dtype == File:
                            if not os.path.isfile(uri.path):
                                raise ParameterValidationError(f"'{mkname(name)}': {uri} is not a regular file")
                        elif dtype == Directory or dtype == MS:
                            if not os.path.isdir(uri.path):
                                raise ParameterValidationError(f"'{mkname(name)}': {uri} is not a directory")
                    inputs[name] = str(uri)
            # else make list
            else:
                # check that files are files and dirs are dirs
                if dtype == List[File]:
                    if not all(
                        os.path.isfile(uri.path) for uri in files if not uri.remote and os.path.exists(uri.path)
                    ):
                        raise ParameterValidationError(f"{mkname(name)}: {value} contains non-files")
                elif dtype == List[Directory] or dtype == List[MS]:
                    if not all(os.path.isdir(uri.path) for uri in files if not uri.remote and os.path.exists(uri.path)):
                        raise ParameterValidationError(f"{mkname(name)}: {value} contains non-directories")
                inputs[name] = list(map(str, files))

    # validate
    try:
        validated = pcls(
            **{name2field[name]: value for name, value in inputs.items() if name in schemas and value is not UNSET}
        )
    except pydantic.ValidationError as exc:
        errors = []
        for err in exc.errors():
            loc = ".".join([field2name.get(x, str(x)) for x in err["loc"]])
            if loc in inputs:
                errors.append(ParameterValidationError(f"{loc} = {inputs[loc]}: {err['msg']}"))
            else:
                errors.append(ParameterValidationError(f"{loc}: {err['msg']}"))
        raise ParameterValidationError(f"{len(errors)} parameter(s) failed validation:", errors)

    validated = {field2name[fld]: value for fld, value in dataclasses.asdict(validated).items()}

    # check choice-type parameters
    for name, value in validated.items():
        if value is None:
            continue
        schema = schemas[name]
        if schema.choices and value not in schema.choices:
            raise ParameterValidationError(f"{mkname(name)}: invalid value '{value}'")
        if schema.element_choices:
            listval = value if isinstance(value, (list, tuple, ListConfig)) else [value]
            for elem in listval:
                if elem not in schema.element_choices:
                    raise ParameterValidationError(f"{mkname(name)}: invalid list element '{elem}'")

    # check for mkdir directives
    if create_dirs:
        for name, value in validated.items():
            schema = schemas[name]
            if schema.is_output and (schema.mkdir or schema.path_policies.mkdir_parent):
                if schema.is_file_type:
                    files = [URI(value)]
                elif schema.is_file_list_type:
                    files = map(URI, value)
                else:
                    continue
                for uri in files:
                    if not uri.remote:
                        path = pathlib.Path(uri.path)
                        # Directory-type outputs
                        if schema.mkdir and (schema._dtype == Directory or schema._dtype == List[Directory]):
                            path.mkdir(parents=True, exist_ok=True)
                        elif schema.path_policies.mkdir_parent:
                            path.parent.mkdir(parents=True, exist_ok=True)

    # add in unresolved values
    validated.update(**unresolved)

    return validated



from omegaconf import DictConfig

from scabha.exceptions import ScabhaBaseException


class ConfigurattError(ScabhaBaseException):
    pass


# paths to search for _include statements
PATH = ["."]

# package version info stored with code dependencies
PACKAGE_VERSION = None

# extensions to look for implicitly, if not supplied
IMPLICIT_EXTENSIONS = (".yml", ".yaml")


# DictConfig doesn't support pop(), so here's a quick replacement
def pop_conf(conf: DictConfig, key: str, default=None):
    value = conf.get(key, default)
    if key in conf:
        del conf[key]
    return value



import glob
import hashlib
import os.path
import pathlib
import sys
from typing import List

import dill as pickle

from .common import PACKAGE_VERSION
from .deps import ConfigDependencies

# path for cache
CACHEDIR = os.environ.get("CONFIGURATT_CACHE_DIR") or os.path.expanduser("~/.cache/configuratt")


def _compute_hash(filelist, extra_keys):
    filelist = list(filelist) + list(extra_keys)
    return hashlib.md5(" ".join(filelist).encode()).hexdigest()


def set_cache_dir(cachedir: str):
    global CACHEDIR
    CACHEDIR = cachedir


def clear_cache(log=None):
    if os.path.isdir(CACHEDIR):
        files = glob.glob(f"{CACHEDIR}/*")
        log and log.info(f"clearing {len(files)} cached config(s) from cache")
    else:
        files = []
        log and log.info("no configs in cache")
    for filename in files:
        try:
            os.unlink(filename)
        except Exception as exc:
            log and log.error(f"failed to remove cached config {filename}: {exc}")
            sys.exit(1)


def load_cache(filelist: List[str], extra_keys=[], verbose=None):
    filehash = _compute_hash(filelist, extra_keys)
    if not os.path.isdir(CACHEDIR):
        if verbose:
            print(f"{CACHEDIR} does not exist")
        return None, None
    filename = os.path.join(CACHEDIR, filehash)
    if not os.path.exists(filename):
        if verbose:
            print(f"hash file {filename} does not exist")
        return None, None
    # check that all configs are older than the cache
    cache_mtime = os.path.getmtime(filename)
    for f in filelist:
        if os.path.getmtime(f) > cache_mtime:
            if verbose:
                print(f"Config {f} is newer than the cache, forcing reload")
            return None, None
    # load cache
    try:
        conf, deps = pickle.load(open(filename, "rb"))
        if not isinstance(deps, ConfigDependencies):
            raise TypeError(f"cached deps object is of type {type(deps)}, expecting ConfigDependencies")
    except Exception as exc:
        print(f"Error loading cached config from {filename}: {exc}. Removing the cache.")
        os.unlink(filename)
        return None, None
    # check that all dependencies are older than the cache
    if deps.have_deps_changed(cache_mtime, verbose=verbose):
        return None, None
    if verbose:
        print(f"Loaded cached config for {' '.join(filelist)} from {filename}")
    return conf, deps


def save_cache(filelist: List[str], conf, deps: ConfigDependencies, extra_keys=[], verbose=False):
    pathlib.Path(CACHEDIR).mkdir(parents=True, exist_ok=True)
    filelist = list(filelist)  # add self to dependencies
    filehash = _compute_hash(filelist, extra_keys)
    filename = os.path.join(CACHEDIR, filehash)
    # add ourselves to dependencies, so that cache is cleared if implementation changes
    deps.add(__file__, version=PACKAGE_VERSION)
    pickle.dump((conf, deps), open(filename, "wb"), 2)
    if verbose:
        print(f"Caching config for {' '.join(filelist)} as {filename}")



import sys
import traceback
from types import TracebackType
from typing import Dict, List, Union
from typing import Optional as _Optional

from omegaconf import DictConfig

logger = None


def set_logger(log):
    global logger
    logger = log


class Error(str):
    """A string that's marked as an error"""

    pass


ALWAYS_REPORT_TRACEBACK = False


class FormattedTraceback(object):
    """This holds the lines of a formatted traceback object."""

    def __init__(self, tb: TracebackType):
        self.lines = [ln.rstrip() for ln in traceback.format_tb(tb)]


class StimelaPendingDeprecationWarning(PendingDeprecationWarning):
    pass


class StimelaDeprecationWarning(DeprecationWarning):
    pass


class ScabhaBaseException(Exception):
    def __init__(
        self,
        message: str,
        nested: _Optional[
            Union[
                Exception,
                TracebackType,
                FormattedTraceback,
                Dict,
                List[Union[Exception, TracebackType, FormattedTraceback, Dict]],
            ]
        ] = None,
        log=None,
        tb=False,
    ):
        """Initializes exception object

        Args:
            message (str): error message
            nested (_Optional[Union[Exception, List[Exception]]]): Nested exception(s). Defaults to None.
            log (logger): if not None, logs the exception to the given logger
        """
        self.message = message
        # include traceback automatically?
        if isinstance(nested, Exception) and (tb or ALWAYS_REPORT_TRACEBACK) and nested is sys.exc_info()[1]:
            nested = [nested, sys.exc_info()[2]]
        if isinstance(nested, (Exception, TracebackType, dict, DictConfig)):
            nested = [nested]
        self.nested = nested or []
        # convert nested tracebacks to formatted ones
        self.nested = [FormattedTraceback(x) if isinstance(x, TracebackType) else x for x in self.nested]

        nested_exc = [str(exc) for exc in self.nested if isinstance(exc, Exception)]
        if nested_exc:
            message = f"{message}: {', '.join(nested_exc)}"
        Exception.__init__(self, message)
        if log is not None:
            if not hasattr(log, "error"):
                log = logger
            if log is not None:
                log.error(message)
        self.logged = log is not None


class ConfigError(ScabhaBaseException):
    pass


class SchemaError(ScabhaBaseException):
    pass


class NestedSchemaError(SchemaError):
    pass


class DefinitionError(ScabhaBaseException):
    pass


class StepValidationError(ScabhaBaseException):
    pass


class AssignmentError(ScabhaBaseException):
    pass


class ParameterValidationError(ScabhaBaseException):
    pass


class SubstitutionError(ScabhaBaseException):
    pass


class UnsetError(ScabhaBaseException):
    pass


class ParserError(ScabhaBaseException):
    pass


class FormulaError(ScabhaBaseException):
    pass


class CyclicSubstitutionError(SubstitutionError):
    def __init__(self, location: List[str], other_location: List[str]):
        self.location = ".".join(location)
        self.other_location = ".".join(other_location)
        super().__init__(f"'{{{self.location}}}' is a cyclic substition")


class SubstitutionErrorList(ScabhaBaseException):
    pass



#  Adapted  from: https://click.palletsprojects.com/en/stable/complex/#lazily-loading-subcommands
import importlib

import click


class LazyGroup(click.Group):
    def __init__(self, *args, lazy_subcommands=None, parent_module=None, **kwargs):
        super().__init__(*args, **kwargs)
        # lazy_subcommands is a map of the form:
        #
        #   {command-name} -> {module-name}.{command-object-name}
        #
        self.lazy_subcommands = lazy_subcommands or {}
        self.parent_module = parent_module

    def list_commands(self, ctx):
        base = super().list_commands(ctx)
        lazy = sorted(self.lazy_subcommands.keys())
        return base + lazy

    def get_command(self, ctx, cmd_name):
        if cmd_name in self.lazy_subcommands:
            return self._lazy_load(cmd_name)
        return super().get_command(ctx, cmd_name)

    def _lazy_load(self, cmd_name):
        # lazily loading a command, first get the module name and attribute name
        import_path = self.lazy_subcommands[cmd_name]
        if isinstance(import_path, click.Command):
            return import_path
        modname, cmd_object_name = import_path.rsplit(".", 1)
        # do the import
        import_prefix = f"{self.parent_module}." if self.parent_module else ""
        mod = importlib.import_module(import_prefix + modname)
        # get the Command object from that module
        cmd_object = getattr(mod, cmd_object_name)
        # check the result to make debugging easier
        if not isinstance(cmd_object, click.Command):
            raise ValueError(f"Lazy loading of {import_path} failed by returning a non-command object")
        return cmd_object



from copy import deepcopy
from dataclasses import fields, is_dataclass


def merge_dataclass_instances(inst1: object, inst2: object):
    """Given two instances of a given dataclass, create a new instance which combines their fields.

    Given dataclass instances, checks that they are the same before creating a new instance which
    combines their values. Values from the second instance are given priority unless they are None.
    This provides a mechanism for combining dataclasses after they have been instantiated as
    opposed to the approach taken by OmegaConf. If the second instance is None, will return a copy
    of the first object.

    Args:
        inst1: Instance of an arbitrary dataclass object.
        inst2: Instance of an arbitrary dataclass object.
    """

    _type = type(inst1)

    if not is_dataclass(_type):
        raise TypeError(f"Object of type {_type} is not a dataclass.")

    if inst2 is None:
        return deepcopy(inst1)

    if not isinstance(inst2, _type):
        raise TypeError(f"Instances are of different types: {_type} and {type(inst2)}.")

    kwargs = {}

    for field in fields(_type):
        field_name = field.name
        val_inst1 = getattr(inst1, field_name)
        val_inst2 = getattr(inst2, field_name)
        kwargs[field_name] = val_inst2 if val_inst2 is not None else val_inst1

    return _type(**kwargs)



# ruff: noqa: E731 - ignore assignment of lambda expressions. TODO(JSKenyon): Fix this
import re
from collections.abc import MutableMapping, MutableSequence, MutableSet
from dataclasses import asdict, dataclass, field, make_dataclass
from typing import Any, Callable, Dict, List, Optional, Union

import click
from omegaconf import MISSING, OmegaConf

from scabha import configuratt
from scabha.exceptions import SchemaError

from .basetypes import EmptyDictDefault, File, is_file_type
from .cargo import _UNSET_DEFAULT, UNSET, Cargo, Parameter, ParameterPolicies


def schema_to_dataclass(io: Dict[str, Parameter], class_name: str, bases=(), post_init: Optional[Callable] = None):
    """
    Converts a scabha schema to a dataclass.
    Each parameter in the schema will correspond to a field. Metadata of fields will contain:
    'help' for info, 'choices' for the choices field, 'parameter' for parameter name

    Args:
        io (Dict[str, Parameter]): dict of parameters
        class_name (str):           name of dataclass
        bases (tuple, optional): [description]. Base classes (passed to bases of make_dataclass())
        post_init (callable, optional): Inserts a __post__init__ method into the class if needed.

    Raises:
        SchemaError: if any parameters are malformed

    Returns:
        dataclass
    """
    field2name = {}
    fields = []
    for name, schema in io.items():
        if type(schema) is not Parameter:
            schema = Parameter(**schema)

        if is_file_type(schema._dtype):
            schema._dtype = str
            schema.dtype = "str"

        # sanitize name: dataclass won't take hyphens or periods
        # so replace with "_" and ensure uniqueness
        fldname = name.replace("-", "_").replace(".", "_")
        while fldname in field2name:
            fldname += "_"
        field2name[fldname] = name

        # form up metadata
        metadata = dict(help=schema.info, parameter=name)
        metadata.update(schema.metadata)
        if schema.choices:
            metadata["choices"] = schema.choices
        if schema.element_choices:
            metadata["element_choices"] = schema.element_choices
        metadata["required"] = required = schema.required

        if required and schema.default is not UNSET:
            raise SchemaError(
                f"Field '{fldname}' is required but specifies a default. This behaviour is unsupported/ambiguous."
            )

        if required:
            fld = field(default=MISSING, metadata=metadata)
        elif isinstance(schema.default, MutableSequence):
            fld = field(default_factory=default_wrapper(list, schema.default), metadata=metadata)
        elif isinstance(schema.default, MutableSet):
            fld = field(default_factory=default_wrapper(set, schema.default), metadata=metadata)
        elif isinstance(schema.default, MutableMapping):
            fld = field(default_factory=default_wrapper(dict, schema.default), metadata=metadata)
        elif schema.default is UNSET:
            fld = field(default=None, metadata=metadata)
        else:
            fld = field(default=schema.default, metadata=metadata)

        fields.append((fldname, schema._dtype, fld))

    namespace = None if post_init is None else dict(__post_init__=post_init)

    return make_dataclass(class_name, fields, bases=bases, namespace=namespace)


def default_wrapper(default_type, default_value):
    def default_factory():
        return default_type(default_value)

    return default_factory


def nested_schema_to_dataclass(nested: Dict[str, Dict], class_name: str, bases=(), section_bases=(), post_init_map={}):
    """Converts a nested schema (sections consisting of parameters) into a dataclass.

    Args:
        nested (Dict[str, Dict]):           schema sections
        class_name (str):                   name of dataclass created
        bases (tuple, optional):            base classes for outer class
        section_bases (tuple, optional):    base class for section classes
        post_init_map (dict, optional):     dict of specific __post_init__ methods to be attached to section classes

    Returns:
        tuple of (dataclass, dict):         resulting dataclass, plus dict of dataclasses for the sections
    """

    # make dataclass based on schema contents
    nested_structure = make_dataclass(
        f"_{class_name}_schemas", [(name, Dict[str, Parameter]) for name in nested.keys()]
    )
    # turn into a structured config
    nested = OmegaConf.unsafe_merge(OmegaConf.structured(nested_structure), nested)
    fields = []

    # convert per-section schemas into dataclasses and make list of fields for outer dataclass
    for section, content in nested.items():
        dcls = schema_to_dataclass(
            content, f"{class_name}_{section}", bases=section_bases, post_init=post_init_map.get(section)
        )

        fields.append((section, dcls, field(default_factory=dcls)))

    # return the outer dataclass
    return make_dataclass(class_name, fields, bases=bases)


_atomic_types = dict(bool=bool, str=str, int=int, float=float)


def _validate_list(text: str, element_type, schema, sep=",", brackets=True):
    if not text:
        if schema.default in (UNSET, _UNSET_DEFAULT):
            return None
        else:
            return schema.default

    if text == "[]":
        return []

    if text[0] == "[" and text[-1] == "]":
        text = text[1:-1]
    elif brackets:
        raise click.BadParameter(f"can't convert to '{schema.dtype}', missing '[]' brackets")

    try:
        return list(element_type(x) for x in text.split(sep))
    except ValueError as e:
        raise click.BadParameter(f"can't convert to '{schema.dtype}'. Underlying exception was: {e}")


def _validate_tuple(text: str, element_types, schema, sep=",", brackets=True):
    if not text:
        if schema.default in (UNSET, _UNSET_DEFAULT):
            return None
        else:
            return schema.default

    if text == "[]":
        return []

    if text[0] == "[" and text[-1] == "]":
        text = text[1:-1]
    elif brackets:
        raise click.BadParameter(f"can't convert to '{schema.dtype}', missing '[]' brackets")

    elems = text.split(sep)
    if len(elems) != len(element_types):
        raise click.BadParameter(f"can't convert to '{schema.dtype}', tuple length mismatch")
    try:
        return tuple(element_type(x) for x, element_type in zip(elems, element_types))
    except ValueError as e:
        raise click.BadParameter(f"can't convert to '{schema.dtype}'. Underlying exception was: {e}")


@dataclass
class Schema(object):
    inputs: Dict[str, Parameter] = EmptyDictDefault()
    outputs: Dict[str, Parameter] = EmptyDictDefault()
    policies: Optional[Dict[str, Any]] = None


def clickify_parameters(schemas: Union[str, Dict[str, Any]], default_policies: Dict[str, Any] = None):
    """
    Create command line parameters from a YAML schema. Uses the click
    package.

    Args:
        schemas (str or Dict): Either the YAML filename from which the parameter schema is loaded,
            containing inputs, outputs and an [optional] policies section,
            or a DictConfig object containing inputs/outputs/policies.
            See https://stimela.readthedocs.io/en/latest/reference/schema_ref.html
        default_policies: default policies applied to the schema, overrides the policies section
            if supplied. See ParameterPolicies in scabha/cargo.py, and
            https://stimela.readthedocs.io/en/latest/reference/policies.html

    Example:
    =======
    The following code defines a simple function that takes two parameters.
    The schema is contained in the file hello/hello.yml. It contains

        inputs:
            count:
                dtype: int
                default: 1
            name:
                dtype: str
                required: true

        outputs:
        {}

    The corresponding python code uses clickify_parameters as
    a decorator:
        import click

        from scabha.schema_utils import clickify_parameters
        from omegaconf import OmegaConf


        @click.command()
        @clickify_parameters('hello/hello.yml')
        def hello(count: int = 1, name: Optional[str] = None):
            for x in range(count):
                click.echo(f"Hello {name}!")
    Returns:
        Nothing
    """

    if type(schemas) is str:
        schemas = OmegaConf.merge(OmegaConf.structured(Schema), OmegaConf.load(schemas))

    # get default policies from argument or schemas
    if default_policies:
        default_policies = OmegaConf.merge(OmegaConf.structured(ParameterPolicies), default_policies)
    elif getattr(schemas, "policies", None):
        default_policies = OmegaConf.merge(OmegaConf.structured(ParameterPolicies), schemas.policies)
    else:
        default_policies = ParameterPolicies()

    decorator_chain = None
    inputs = Cargo.flatten_schemas({}, getattr(schemas, "inputs", {}), "inputs")
    outputs = Cargo.flatten_schemas({}, getattr(schemas, "outputs", {}), "outputs")
    for io in inputs, outputs:
        for name, schema in io.items():
            # skip outputs, unless they're named outputs
            if io is outputs and not (schema.is_file_type and not schema.implicit):
                continue

            # sometimes required to convert ParameterPolicies object to dict
            if isinstance(schema.policies, ParameterPolicies):
                merge_policies = {k: v for k, v in asdict(schema.policies).items() if v is not None}
            else:
                merge_policies = {k: v for k, v in schema.policies.items() if v is not None}
            # None should not take precedence in the merge
            policies = OmegaConf.merge(default_policies, merge_policies)

            # impose default repeat policy of using a single argument for a list, i.e. X1,X2,X3
            if policies.repeat is None:
                policies.repeat = ","

            name = name.replace("_", "-").replace(".", "-")
            optname = f"--{name}"
            dtype = schema.dtype
            if type(dtype) is str:
                dtype = dtype.strip()
            validator = None
            multiple = False
            nargs = 1
            # process optional
            optional_match = re.fullmatch(r"Optional\[(.*)\]", dtype)
            if optional_match:
                str_dtype = dtype = optional_match.group(1).strip()
            else:
                str_dtype = str(dtype)
            metavar = schema.metavar or str_dtype
            is_unset = schema.default in (UNSET, _UNSET_DEFAULT)
            kwargs = dict()
            if not is_unset and not schema.suppress_cli_default and nargs != -1:
                kwargs["default"] = schema.default

            # sort out option type. Atomic type?
            if dtype in _atomic_types:
                dtype = _atomic_types[dtype]
                if dtype is bool:
                    if getattr(policies, "is_flag", None):
                        kwargs["is_flag"] = True
                    elif getattr(policies, "explicit_flag"):
                        # click default behavior for type 'bool'
                        pass
                    else:
                        # leave this as the default behavior for backwards compatibility
                        # but it really should be the first clause
                        optname = f"{optname}/--no-{name}"
            # file type? NB: URI not included deliberately -- this becomes a str in the else: clause below
            elif dtype in ("MS", "File", "Directory"):
                dtype = click.Path(exists=(io is schemas.inputs))
            else:
                list_match = re.fullmatch(r"List\[(.*)\]", dtype)
                tuple_match = re.fullmatch(r"Tuple\[(.*)\]", dtype)
                # List[x] type? Add validation callback to convert elements
                if list_match:
                    elem_type = _atomic_types.get(list_match.group(1).strip(), str)
                    if policies.repeat == "list":
                        if not policies.positional:
                            raise SchemaError(
                                f"click parameter '{name}': repeat=list policy is only supported for "
                                f"positional=true parameters"
                            )
                        nargs = -1
                        dtype = elem_type
                        metavar = schema.metavar or f"{elem_type.__name__} ..."
                    elif policies.repeat == "repeat":
                        multiple = True
                        dtype = elem_type
                        metavar = schema.metavar or f"{elem_type.__name__}"
                        # multiple options have a click default of () not None
                        if "default" not in kwargs:
                            kwargs["default"] = None
                    elif policies.repeat == "[]":  # else assume [X,Y] or X,Y syntax
                        dtype = str
                        validator = lambda ctx, param, value, etype=dtype, schema=schema, _type=elem_type: (
                            _validate_list(value, element_type=_type, schema=schema, brackets=False)
                        )
                        metavar = schema.metavar or f"{elem_type.__name__},{elem_type.__name__},..."
                    elif policies.repeat is not None:  # assume XrepY syntax
                        dtype = str
                        sep = policies.repeat
                        validator = lambda ctx, param, value, etype=dtype, schema=schema, _type=elem_type: (
                            _validate_list(value, element_type=_type, schema=schema, sep=sep, brackets=False)
                        )
                        metavar = schema.metavar or f"{elem_type.__name__}{sep}{elem_type.__name__}{sep}..."
                    else:
                        raise SchemaError(f"list-type parameter '{name}' does not have a repeat policy set")
                elif tuple_match:
                    elem_types = tuple(_atomic_types.get(t.strip(), str) for t in tuple_match.group(1).split(","))
                    if policies.repeat == "list":
                        nargs = len(elem_types)
                        dtype = elem_types
                        metavar = schema.metavar or " ".join(t.__name__ for t in elem_types)
                    elif policies.repeat == "repeat":
                        raise SchemaError(f"tuple-type parameter '{name}' has unsupported repeat policy 'repeat'")
                    elif policies.repeat == "[]":  # else assume [X,Y] or X,Y syntax
                        dtype = str
                        metavar = schema.metavar or ",".join((t.__name__ for t in elem_types))
                        validator = lambda ctx, param, value, etype=dtype, schema=schema, _type=elem_types: (
                            _validate_tuple(value, element_types=_type, schema=schema, brackets=False)
                        )
                    elif policies.repeat is not None:  # assume XrepY syntax
                        dtype = str
                        metavar = schema.metavar or policies.repeat.join((t.__name__ for t in elem_types))
                        validator = lambda ctx, param, value, etype=dtype, schema=schema, _type=elem_types: (
                            _validate_tuple(
                                value, element_types=_type, schema=schema, sep=policies.repeat, brackets=False
                            )
                        )
                    else:
                        raise SchemaError(f"tuple-type parameter '{name}' does not have a repeat policy set")
                else:
                    # anything else will be just a string
                    dtype = str

            # choices?
            if schema.choices:
                dtype = click.Choice(schema.choices)

            # aliases?
            optnames = [optname]
            if schema.abbreviation:
                optnames.append(f"-{schema.abbreviation}")

            if policies.positional:
                kwargs.update(type=dtype, callback=validator, required=schema.required, nargs=nargs, metavar=metavar)
                deco = click.argument(name, **kwargs)
            else:
                kwargs.update(
                    type=dtype,
                    callback=validator,
                    required=schema.required,
                    multiple=multiple,
                    metavar=metavar,
                    help=schema.info,
                )
                deco = click.option(*optnames, **kwargs)
            if decorator_chain is None:
                decorator_chain = deco
            else:
                decorator_chain = lambda x, deco=deco, chain=decorator_chain: chain(deco(x))

    return decorator_chain or (lambda x: x)


@dataclass
class SchemaSpec:
    inputs: Dict[str, Parameter]
    outputs: Dict[str, Parameter]
    libs: Optional[Dict[str, Any]] = EmptyDictDefault()
    info: Optional[str] = None


def paramfile_loader(
    paramfiles: Union[File, List[File]], sources: Union[File, List[File]] = [], schema_spec=None, use_cache=False
) -> Dict:
    """Load a scabha-style parameter defintion using.

    Args:
        paramfiles (List[File]): Name of parameter definition files
        sources (List[Dict], optional): Parameter definition dependencies
        (a.k.a files specified via_include)

    Returns:
        Dict: Schema object
    """
    args_defn = OmegaConf.structured(schema_spec or SchemaSpec)
    if isinstance(paramfiles, File):
        paramfiles = [paramfiles]

    if isinstance(sources, File):
        sources = [sources]

    srcs = []
    for src in sources:
        if not src.EXISTS:
            raise FileNotFoundError(f"Source file for either of {paramfiles} could not be found at {src.PATH}")
        srcs.append(configuratt.load(src, use_cache=use_cache)[0])

    struct_args, _ = configuratt.load_nested(paramfiles, structured=args_defn, use_sources=srcs, use_cache=use_cache)

    return OmegaConf.create(struct_args)



import os
import os.path
import stat


def which(binary, extra_paths=None):
    """Equivalent of shell which command. Returns full path to executable, or None if not found"""
    for path in (extra_paths or []) + os.environ["PATH"].split(os.pathsep):
        fullpath = os.path.join(path, binary)
        if os.path.isfile(fullpath) and os.stat(fullpath).st_mode & stat.S_IXUSR:
            return fullpath
    return None



import fnmatch
from typing import Dict, List, Union

from omegaconf.omegaconf import DictConfig

from .common import ConfigurattError, pop_conf


def _lookup_nameseq(name_seq: List[str], source_dict: Dict):
    """Internal helper: looks up nested item ('a', 'b', 'c') in a nested dict

    Parameters
    ----------
    name_seq : List[str]
        sequence of keys to look up
    source_dict : Dict
        nested dict

    Returns
    -------
    Any
        value if found, else None
    """
    source = source_dict
    names = list(name_seq)
    while names:
        source = source.get(names.pop(0), None)
        if source is None:
            return None
    return source


def _lookup_name(name: str, *sources: List[Dict]):
    """Internal helper: looks up a nested item ("a.b.c") in a list of dicts

    Parameters
    ----------
    name : str
        section name to look up, e.g. "a.b.c"

    Returns
    -------
    Any
        first matching item found

    Raises
    ------
    NameError
        if matching item is not found
    """
    name_seq = name.split(".")
    for source in sources:
        result = _lookup_nameseq(name_seq, source)
        if result is not None:
            return result
    raise ConfigurattError(f"unknown key {name}")


def _flatten_subsections(conf, depth: int = 1, sep: str = "__"):
    """Recursively flattens subsections in a DictConfig (modifying in place)
    A structure such as
        a:
            b: 1
            c: 2
    Becomes
        a__b: 1
        a__c: 2

    Args:
        conf (DictConfig): config to flatten
        depth (int):       depth to which to flatten. Default is 1 level.
        sep (str):         separator to use, default is "__"
    """
    subsections = [(key, value) for key, value in conf.items() if isinstance(value, DictConfig)]
    for name, subsection in subsections:
        pop_conf(conf, name)
        if depth > 1:
            _flatten_subsections(subsection, depth - 1, sep)
        for key, value in subsection.items():
            conf[f"{name}{sep}{key}"] = value


def _scrub_subsections(conf: DictConfig, scrubs: Union[str, List[str]]):
    """
    Scrubs named subsections from a config.

    Args:
        conf (DictConfig): config to scrub
        scrubs (Union[str, List[str]]): sections to remove (can include dots to remove nested sections)
    """
    if isinstance(scrubs, str):
        scrubs = [scrubs]

    for scrub in scrubs:
        if "." in scrub:
            name, remainder = scrub.split(".", 1)
        else:
            name, remainder = scrub, None
        # apply name as pattern
        is_pattern = "*" in name or "?" in name
        matches = fnmatch.filter(conf.keys(), name)
        if not matches:
            # if no matches to pattern, it's ok, otherwise raise error
            if is_pattern:
                return
            raise ConfigurattError(f"no entry matching '{name}'")
        # recurse into or remove matching entries
        for key in matches:
            if remainder:
                subconf = conf[key]
                if type(subconf) is DictConfig:
                    _scrub_subsections(subconf, remainder)
                elif not is_pattern:
                    raise ConfigurattError(f"'{name}' does not refer to a subsection")
            else:
                del conf[key]



import datetime
import fnmatch
import hashlib
import importlib
import os.path
import subprocess
from dataclasses import dataclass
from shutil import which
from typing import List, Optional

from omegaconf.omegaconf import OmegaConf


@dataclass
class FailRecord(object):
    filename: str
    origin: Optional[str] = None
    modulename: Optional[str] = None
    fname: Optional[str] = None
    warn: bool = True


@dataclass
class RequirementRecord(object):
    location: str
    requires: str
    filename: str
    optional: bool = False


class ConfigDependencies(object):
    _git_cache = {}

    def __init__(self):
        self.deps = OmegaConf.create()
        self.fails = OmegaConf.create()
        self._git = which("git")
        # self.provides = OmegaConf.create()
        # self.requires = OmegaConf.create()

    def add(self, filename: str, origin: Optional[str] = None, missing=False, **extra_attrs):
        """Adds a file to the set of dependencies

        Args:
            filename (str): filename
            origin (str or None): if not None, marks dependency as originating from another dependency
            missing (bool, optional): If True, marks depndency as missing. Defaults to False.
        """
        filename = os.path.abspath(filename)
        if filename in self.deps:
            return
        depinfo = OmegaConf.create()
        depinfo.update(**extra_attrs)
        if origin is not None:
            depinfo.origin = origin
        else:
            if missing or not os.path.exists(filename):
                depinfo.mtime = 0
                depinfo.mtime_str = "n/a"
                self.deps[filename] = depinfo
                return
            # get mtime and hash
            depinfo.mtime = os.path.getmtime(filename)
            depinfo.mtime_str = datetime.datetime.fromtimestamp(depinfo.mtime).strftime("%c")
            if not os.path.isdir(filename):
                depinfo.md5hash = hashlib.md5(open(filename, "rb").read()).hexdigest()
            # add git info
            dirname = os.path.realpath(filename)
            if not os.path.isdir(dirname):
                dirname = os.path.dirname(dirname)
            gitinfo = self._get_git_info(dirname)
            if gitinfo:
                depinfo.git = gitinfo
        self.deps[filename] = depinfo

    def add_fail(self, fail: FailRecord):
        self.fails[fail.filename] = OmegaConf.structured(fail)

    def replace(self, globs: List[str], dirname: str, **extra_attrs):
        remove = set()
        for glob in globs:
            remove.update(fnmatch.filter(self.deps, glob))
        if remove:
            for name in remove:
                self.deps[name] = OmegaConf.create(dict(origin=dirname))
        # add directory
        if dirname not in self.deps:
            self.add(dirname, **extra_attrs)

    def update(self, other):
        for name in other.deps:
            if name not in self.deps:
                self.deps[name] = other.deps[name]
        self.fails = OmegaConf.unsafe_merge(self.fails, other.fails)
        # self.provides = OmegaConf.unsafe_merge(self.provides, other.provides)
        # self.requires = OmegaConf.unsafe_merge(self.requires, other.requires)

    def save(self, filename):
        OmegaConf.save(self.deps, filename)

    def _get_git_info(self, dirname: str):
        """Returns git info structure for a directory, or None if not under git control

        Args:
            dirname (str): path

        Returns:
            DictConfig: directory info
        """
        # check cache first
        if dirname in self._git_cache:
            return self._git_cache[dirname]
        if not self._git:
            return None
        try:
            branches = subprocess.check_output(
                "git -c color.ui=never branch -a -v -v".split(), cwd=dirname, stderr=subprocess.DEVNULL
            )
        except subprocess.CalledProcessError:
            self._git_cache[dirname] = None
            return None
        # use git to get the info
        gitinfo = OmegaConf.create()
        for line in branches.decode().split("\n"):
            line = line.strip()
            if line.startswith("*"):
                gitinfo.branch = line[1:].strip().replace("${", "\\${")
                break
        # get description
        try:
            describe = subprocess.check_output("git describe --abbrev=16 --always --long --all".split(), cwd=dirname)
            gitinfo.describe = describe.decode().strip()
        except subprocess.CalledProcessError:
            pass
        # get remote info
        try:
            remotes = subprocess.check_output("git remote -v".split(), cwd=dirname)
            gitinfo.remotes = remotes.decode().strip().split("\n")
        except subprocess.CalledProcessError:
            pass

        self._git_cache[dirname] = gitinfo
        return gitinfo

    def get_description(self):
        desc = {}
        for filename, attrs in self.deps.items():
            attrs_items = attrs.items() if attrs else []
            attrs_str = [
                f"mtime: {datetime.datetime.fromtimestamp(value).strftime('%c')}"
                if attr == "mtime"
                else f"{attr}: {value}"
                for attr, value in attrs_items
            ]
            desc[filename] = attrs_str
        return desc

    def have_deps_changed(self, mtime, verbose=False):
        # check that all dependencies are older than the cache
        for f in self.deps.keys():
            if not os.path.exists(f):
                if verbose:
                    print(f"Dependency {f} doesn't exist, forcing reload")
                return True
            if os.path.getmtime(f) > mtime:
                if verbose:
                    print(f"Dependency {f} is newer than the cache, forcing reload")
                return True
        # check that previously failing includes are not now succeeding (because that's also reason to reload cache)
        for filename, dep in self.fails.items():
            if dep.modulename:
                try:
                    mod = importlib.import_module(dep.modulename)
                    fname = os.path.join(os.path.dirname(mod.__file__), dep.fname)
                    if os.path.exists(fname):
                        return True
                except ImportError:
                    pass
            elif not dep.missing_parent:
                if os.path.exists(filename):
                    return True
        return False

    # def add_provision_record(self, loc, filename):
    #     if loc not in self.provides:
    #         self.provides[loc] = []
    #     if filename not in self.provides[loc]:
    #         self.provides[loc].append(filename)

    # def scan_requirements(self, conf: DictConfig, location: Optional[str], filename: str):
    #     # build requirements map first using recursive helper
    #     def _scan(conf, loc, filename):
    #         if isinstance(conf, DictConfig):
    #             # add to requirements map, if this has requirements
    #             for optional, keyword in (False, "_requires"), (True, "_contingent"):
    #                 reqs = pop_conf(conf, keyword, [])
    #                 reqs = [reqs] if type(reqs) is str else reqs
    #                 # make list of unmet requirements
    #                 reqs = [req for req in reqs if (loc if req == "_base" else req) not in self.provides]
    #                 # save remaining unresolved reqs
    #                 if reqs:
    #                     if loc not in self.requires:
    #                         self.requires[loc] = OmegaConf.create()
    #                     for req in reqs:
    #                         reqloc = loc if req == "_base" else req
    #                         self.requires[loc][req] = RequirementRecord(loc, reqloc, filename, optional=optional)
    #             # If all resolved, add to provision record
    #             if loc not in self.requires and loc not in self.provides:
    #                 self.provides[loc] = filename
    #             # recurse into content
    #             for name, value in conf.items_ex(resolve=False):
    #                 _scan(value, f"{loc}.{name}" if loc else name, filename)

    #     _scan(conf, location or "", filename)

    # def check_requirements(self, conf: DictConfig, strict=True):
    #     """Checks remaining unmet requirements"""
    #     unmet = []
    #     optional = {}
    #     for loc, reqs in self.requires.items_ex(resolve=False):
    #         for name, req in reqs.items_ex(resolve=False):
    #             if req.requires not in self.provides:
    #                 if not strict or req.optional:
    #                     optional[loc, name] = req
    #                 else:
    #                     unmet.append(ConfigurattError(
    #                         f"requirement '{name}' not met for section '{loc}' in {req.filename}"
    #                     ))
    #     if unmet:
    #         raise ConfigurattError("configuration has missing requirements", nested=unmet)

    #     for loc, _ in optional.keys():
    #         section = conf
    #         loc_elems = loc.split(".")
    #         try:
    #             for loc_elem in loc_elems[:-1]:
    #                 section = section[loc_elem]
    #             del section[loc_elem[-1]]
    #         except KeyError as exc:
    #             pass

    #     return optional



import importlib
import os.path
import re
import uuid
from collections.abc import Sequence
from dataclasses import make_dataclass
from typing import Any, Callable, List, Optional, Union

from omegaconf.errors import OmegaConfBaseException
from omegaconf.omegaconf import DictConfig, ListConfig, OmegaConf
from yaml.error import YAMLError

from .cache import load_cache, save_cache
from .common import IMPLICIT_EXTENSIONS, PATH, ConfigurattError, pop_conf
from .deps import ConfigDependencies, FailRecord
from .helpers import _lookup_name, _scrub_subsections


def load(
    path: str,
    use_sources: Optional[List[DictConfig]] = [],
    name: Optional[str] = None,
    location: Optional[str] = None,
    includes: bool = True,
    selfrefs: bool = True,
    include_path: str = None,
    use_cache: bool = True,
    no_toplevel_cache=False,
    include_stack=[],
    verbose: bool = False,
):
    """Loads config file, using a previously loaded config to resolve _use references.

    Args:
        path (str): path to config file
        use_sources (Optional[List[DictConfig]]): list of existing configs to be used to resolve "_use" references,
                or None to disable
        name (Optional[str]): name of this config file, used for error messages
        location (Optional[str]): location where this config is being loaded (if not at root level)
        includes (bool, optional): If True (default), "_include" references will be processed
        selfrefs (bool, optional): If False, "_use" references will only be looked up in existing config.
            If True (default), they'll also be looked up within the loaded config.
        include_stack: list of paths which have been included. Used to catch recursive includes.
        include_path (str, optional):
            if set, path to each config file will be included in the section as element 'include_path'

    Returns:
        Tuple of (conf, dependencies)
            conf (DictConfig): config object
            dependencies (ConfigDependencies): filenames that were _included
    """
    use_toplevel_cache = use_cache and not no_toplevel_cache
    conf, dependencies = load_cache((path,), verbose=verbose) if use_toplevel_cache else (None, None)

    if conf is None:
        # create self:xxx resolver
        self_namespace = dict(path=path, dirname=os.path.dirname(path), basename=os.path.basename(path))

        def self_namespace_resolver(arg):
            if arg in self_namespace:
                return self_namespace[arg]
            raise KeyError(f"invalid '${{self:arg}}' substitution in {path}")

        OmegaConf.register_new_resolver("self", self_namespace_resolver)
        try:
            subconf = OmegaConf.load(path)
            # force resolution of interpolations at this point (otherwise they happen lazily)
            resolved = OmegaConf.to_container(subconf, resolve=True)
            subconf = OmegaConf.create(resolved)
        finally:
            OmegaConf.clear_resolver("self")

        name = name or os.path.basename(path)
        dependencies = ConfigDependencies()
        dependencies.add(path)
        # include ourself into sources, if _use is in effect, and we've enabled selfrefs
        if use_sources is not None and selfrefs:
            use_sources = [subconf] + list(use_sources)
        conf, deps = resolve_config_refs(
            subconf,
            pathname=path,
            location=location,
            name=name,
            includes=includes,
            use_cache=use_cache,
            use_sources=use_sources,
            include_path=include_path,
            include_stack=include_stack + [path],
        )
        # update overall dependencies
        dependencies.update(deps)

        # # check for missing requirements
        # dependencies.scan_requirements(conf, location, path)

        if use_cache:
            save_cache((path,), conf, dependencies, verbose=verbose)

    return conf, dependencies


def load_nested(
    filelist: List[str],
    structured: Optional[DictConfig] = None,
    typeinfo=None,
    use_sources: Optional[List[DictConfig]] = [],
    location: Optional[str] = None,
    nameattr: Union[Callable, str, None] = None,
    config_class: Optional[str] = None,
    include_path: Optional[str] = None,
    use_cache: bool = True,
    verbose: bool = False,
):
    """Builds nested configuration from a set of YAML files corresponding to sub-sections

    Parameters
    ----------
    conf : OmegaConf object
        root OmegaConf object to merge content into
    filelist : List[str]
        list of subsection config files to load
    schema : Optional[DictConfig]
        schema to be applied to each file, if any
    use_sources : Optional[List[DictConfig]]
        list of existing configs to be used to resolve "_use" references, or None to disable
    location : Optional[str]
        if set, contents of files are being loaded under 'location.subsection_name'. If not set, then 'subsection_name'
        is being loaded at root level. This is used for correctly formatting error messages and such.
    nameattr : Union[Callable, str, None]
        if None, subsection_name will be taken from the basename of the file. If set to a string such as 'name', will
        set subsection_name from that field in the subsection config. If callable, will be called with the subsection
        config object as a single argument, and must return the subsection name
    config_class : Optional[str]
        name of config dataclass to form (when using typeinfo), if None, then generated automatically
    include_path : Optional[str]
        if set, path to each config file will be included in the section as element 'include_path'

    Returns
    -------
        Tuple of (conf, dependencies)
            conf (DictConfig): config object
            dependencies (set): set of filenames that were _included

    Raises
    ------
    NameError
        If subsection name is not resolved
    """
    section_content, dependencies = load_cache(filelist, verbose=verbose) if use_cache else (None, None)

    if section_content is None:
        section_content = {}  # OmegaConf.create()
        dependencies = ConfigDependencies()

        for path in filelist:
            # load file
            subconf, deps = load(path, location=location, use_sources=use_sources, include_path=include_path)
            dependencies.update(deps)
            if include_path:
                subconf[include_path] = path

            # figure out section name
            if nameattr is None:
                name = os.path.splitext(os.path.basename(path))[0]
            elif callable(nameattr):
                name = nameattr(subconf)
            elif nameattr in subconf:
                name = subconf.get(nameattr)
            else:
                raise NameError(f"{path} does not contain a '{nameattr}' field")

            # apply schema
            if structured is not None:
                try:
                    subconf = OmegaConf.merge(structured, subconf)
                except (OmegaConfBaseException, YAMLError) as exc:
                    raise ConfigurattError(f"schema error in {path}: {exc}")

            section_content[name] = subconf

        if structured is None and typeinfo is not None:
            if config_class is None:
                config_class = "ConfigClass_" + uuid.uuid4().hex
            fields = [(name, typeinfo) for name in section_content.keys()]
            datacls = make_dataclass(config_class, fields)
            # datacls.__module__ == __name__  # for pickling
            structured = OmegaConf.structured(datacls)
            section_content = OmegaConf.merge(structured, section_content)

        if use_cache:
            save_cache(filelist, section_content, dependencies, verbose=verbose)

    return section_content, dependencies


def resolve_config_refs(
    conf,
    pathname: str,
    location: str,
    name: str,
    includes: bool,
    use_sources: Optional[List[DictConfig]],
    use_cache=True,
    include_path: Optional[str] = None,
    include_stack=[],
):
    """Resolves cross-references ("_use" and "_include" statements) in config object

    Parameters
    ----------
    conf : OmegaConf object
        input configuration object
    pathname : str
        full path to this config (directory component of that is used for _includes)
    location : str
        location of this configuration section, used for messages
    name : str
        name of this configuration file, used for messages
    includes : bool
        If True, "_include" references will be processed
    use_sources : optional list of OmegaConf objects
        one or more config object(s) in which to look up "_use" references. None to disable _use statements
    include_path (str, optional):
        if set, path to each config file will be included in the section as element 'include_path'
    include_stack (list, optional):
        stack of files from which this one was included. Used to catch recursion.

    Returns
    -------
    Tuple of (conf, dependencies)
    conf : OmegaConf object
        This may be a new object if a _use key was resolved, or it may be the existing object
    dependencies : ConfigDependencies
        Set of filenames that were _included

    Raises
    ------
    ConfigurattError
        If a _use or _include directive is malformed
    """
    errloc = f"config error at {location or 'top level'} in {name}"
    dependencies = ConfigDependencies()
    # self-referencing enabled if first source is ourselves
    selfrefs = use_sources and conf is use_sources[0]

    if isinstance(conf, DictConfig):
        ## NB: perhaps have _use and _include take effect at the point they're inserted?
        ## also add an _all statement to insert a section into all section that follow
        # since _use and _include statements can be nested, keep on processing until all are resolved
        updated = True
        recurse = 0

        while updated:
            updated = False
            # check for infinite recursion
            recurse += 1
            if recurse > 20:
                raise ConfigurattError(f"{errloc}: recursion limit exceeded, check your _use and _include statements")
            # handle _include/_include_post entries
            if includes:
                # helper function: process includes recursively
                def process_include_directive(include_files: List[str], keyword: str, directive: Any, subpath=None):
                    if isinstance(directive, str):
                        include_files.append(directive if subpath is None else f"{subpath}/{directive}")
                    elif isinstance(directive, (tuple, list, ListConfig)):
                        for dir1 in directive:
                            process_include_directive(include_files, keyword, dir1, subpath)
                    elif isinstance(directive, DictConfig):
                        for key, value in directive.items_ex():
                            process_include_directive(
                                include_files, keyword, value, subpath=key if subpath is None else f"{subpath}/{key}"
                            )
                    else:
                        raise ConfigurattError(f"{errloc}: {keyword} contains invalid entry of type {type(directive)}")

                # helper function: load list of _include or _include_post files, returns accumulated DictConfig
                def load_include_files(keyword):
                    # pop include directive, return if None
                    include_directive = pop_conf(conf, keyword, None)
                    if include_directive is None:
                        return None
                    # get corresponding _scrub or _scrub_post directive
                    scrub = pop_conf(conf, keyword.replace("include", "scrub"), None)
                    if isinstance(scrub, str):
                        scrub = [scrub]

                    include_files = []
                    process_include_directive(include_files, keyword, include_directive)

                    accum_incl_conf = OmegaConf.create()

                    # load includes
                    for incl in include_files:
                        if not incl:
                            raise ConfigurattError(f"{errloc}: empty {keyword} specifier")
                        # check for [flags] at end of specifier
                        match = re.match(r"^(.*)\[(.*)\]$", incl)
                        if match:
                            incl = match.group(1)
                            flags = set([x.strip().lower() for x in match.group(2).split(",")])
                            warn = "warn" in flags
                            optional = "optional" in flags
                        else:
                            flags = {}
                            warn = optional = False

                        # helper function -- finds given include file (including trying an implicit .yml or .yaml
                        # extension) returns full name of file if found, else return None if include is optional,
                        # else adds fail record and raises exception. If opt=True, this is stronger than optional
                        # (no warnings raised)
                        def find_include_file(path: str, opt: bool = False):
                            # if path already has an extension, only try the pathname itself
                            if os.path.splitext(path)[1]:
                                paths = [path]
                            # else try the pathname itself, plus implicit extensions
                            else:
                                paths = [path] + [path + ext for ext in IMPLICIT_EXTENSIONS]
                            # now try all of them and return a matching one if found
                            for path in paths:
                                if os.path.isfile(path):
                                    return path
                            # end of loop with no matching files? Raise error
                            else:
                                if opt:
                                    return None
                                elif optional:
                                    dependencies.add_fail(FailRecord(path, pathname, warn=warn))
                                    if warn:
                                        print(f"Warning: unable to find optional include {path}")
                                    return None
                                raise ConfigurattError(f"{errloc}: {keyword} {path} does not exist")

                        # check for (location)filename.yaml or (location)/filename.yaml style
                        match = re.match(r"^\((.+)\)/?(.+)$", incl)
                        if match:
                            modulename, filename = match.groups()
                            if modulename.startswith("."):
                                filename = os.path.join(os.path.dirname(pathname), modulename, filename)
                                filename = find_include_file(filename)
                                if filename is None:
                                    continue
                            else:
                                try:
                                    mod = importlib.import_module(modulename)
                                except ImportError as exc:
                                    if optional:
                                        dependencies.add_fail(
                                            FailRecord(incl, pathname, modulename=modulename, fname=filename, warn=warn)
                                        )
                                        if warn:
                                            print(f"Warning: unable to import module for optional include {incl}")
                                        continue
                                    raise ConfigurattError(
                                        f"{errloc}: {keyword} {incl}: can't import {modulename} ({exc})"
                                    )
                                if mod.__file__ is not None:
                                    path = os.path.dirname(mod.__file__)
                                else:
                                    path = getattr(mod, "__path__", None)
                                    if path is None:
                                        if optional:
                                            dependencies.add_fail(
                                                FailRecord(
                                                    incl, pathname, modulename=modulename, fname=filename, warn=warn
                                                )
                                            )
                                            if warn:
                                                print(
                                                    f"Warning: unable to resolve path for optional include {incl}, "
                                                    f"does {modulename} contain __init__.py?"
                                                )
                                            continue
                                        raise ConfigurattError(
                                            f"{errloc}: {keyword} {incl}: can't resolve path for {modulename}, does "
                                            f"it contain __init__.py?"
                                        )
                                    path = path[0]

                                filename = find_include_file(os.path.join(path, filename))
                                if filename is None:
                                    continue
                        # absolute path -- one candidate
                        elif os.path.isabs(incl):
                            filename = find_include_file(incl)
                            if filename is None:
                                continue
                        # relative path -- scan PATH for candidates
                        else:
                            paths = [".", os.path.dirname(pathname)] + PATH
                            candidates = [os.path.join(p, incl) for p in paths]
                            for filename in candidates:
                                filename = find_include_file(filename, opt=True)
                                if filename is not None:
                                    break
                            # none found in candidates -- process error
                            else:
                                if optional:
                                    dependencies.add_fail(FailRecord(incl, pathname, warn=warn))
                                    if warn:
                                        print(f"Warning: unable to find optional include {incl}")
                                    continue
                                raise ConfigurattError(f"{errloc}: {keyword} {incl} not found in {':'.join(paths)}")

                        # check for recursion
                        for path in include_stack:
                            if os.path.samefile(path, filename):
                                raise ConfigurattError(f"{errloc}: {filename} is included recursively")
                        # load included file
                        incl_conf, deps = load(
                            filename,
                            location=location,
                            name=f"{filename}, included from {name}",
                            includes=True,
                            include_stack=include_stack,
                            use_cache=use_cache,
                            use_sources=None,
                        )  # do not expand _use statements in included files, this is done below

                        dependencies.update(deps)
                        if include_path is not None:
                            incl_conf[include_path] = filename

                        # accumulate included config so that later includes override earlier ones
                        accum_incl_conf = OmegaConf.unsafe_merge(accum_incl_conf, incl_conf)

                    if scrub:
                        try:
                            _scrub_subsections(accum_incl_conf, scrub)
                        except ConfigurattError as exc:
                            raise ConfigurattError(f"{errloc}: error scrubbing {', '.join(scrub)}", exc)

                    return accum_incl_conf

                accum_pre = load_include_files("_include")
                accum_post = load_include_files("_include_post")

                # merge: our section overrides anything that has been included
                conf = OmegaConf.unsafe_merge(accum_pre or {}, conf, accum_post or {})
                if accum_pre or accum_post:
                    updated = True
                if selfrefs:
                    use_sources[0] = conf

            # handle _use entries
            if use_sources is not None:

                def load_use_sections(keyword):
                    merge_sections = pop_conf(conf, keyword, None)
                    if merge_sections is None:
                        return None
                    scrub = pop_conf(conf, keyword.replace("use", "scrub"), None)
                    if type(merge_sections) is str:
                        merge_sections = [merge_sections]
                    elif not isinstance(merge_sections, Sequence):
                        raise TypeError(f"invalid {name}.{keyword} directive of type {type(merge_sections)}")
                    if len(merge_sections):
                        # convert to actual sections
                        merge_sections = [_lookup_name(name, *use_sources) for name in merge_sections]
                        # merge them all together
                        base = merge_sections[0].copy()
                        base.merge_with(*merge_sections[1:])
                        # resolve references before flattening
                        base, deps = resolve_config_refs(
                            base,
                            pathname=pathname,
                            name=name,
                            location=f"{location}.{keyword}" if location else keyword,
                            includes=includes,
                            use_sources=use_sources,
                            use_cache=use_cache,
                            include_path=include_path,
                        )
                        dependencies.update(deps)
                        if scrub:
                            try:
                                _scrub_subsections(base, scrub)
                            except ConfigurattError as exc:
                                raise ConfigurattError(f"{errloc}: error scrubbing {', '.join(scrub)}", exc)
                        return base
                    return None

                base = load_use_sections("_use")
                if base is not None:
                    base.merge_with(conf)
                    conf = base
                post = load_use_sections("_use_post")
                if post is not None:
                    conf.merge_with(post)

                if selfrefs:
                    use_sources[0] = conf

        # recurse into content
        for key, value in conf.items_ex(resolve=False):
            if isinstance(value, (DictConfig, ListConfig)):
                value1, deps = resolve_config_refs(
                    value,
                    pathname=pathname,
                    name=name,
                    location=f"{location}.{key}" if location else key,
                    includes=includes,
                    include_stack=include_stack,
                    use_sources=use_sources,
                    use_cache=use_cache,
                    include_path=include_path,
                )
                dependencies.update(deps)
                # reassigning is expensive, so only do it if there was an actual change
                if value1 is not value:
                    conf[key] = value1

    # recurse into lists
    elif isinstance(conf, ListConfig):
        # recurse in
        for i, value in enumerate(conf._iter_ex(resolve=False)):
            if isinstance(value, (DictConfig, ListConfig)):
                value1, deps = resolve_config_refs(
                    value,
                    pathname=pathname,
                    name=name,
                    location=f"{location or ''}[{i}]",
                    includes=includes,
                    include_stack=include_stack,
                    use_sources=use_sources,
                    use_cache=use_cache,
                    include_path=include_path,
                )
                dependencies.update(deps)
                if value1 is not value:
                    conf[i] = value

    return conf, dependencies



# ruff: noqa: E741 - ignore ambiguous variable name.
import dataclasses
import fnmatch
import glob
import operator
import os.path
import typing
from functools import reduce
from typing import Any, Dict, List

import pyparsing as pp
from omegaconf import DictConfig, ListConfig

from .basetypes import UNSET, Unresolved
from .exceptions import FormulaError, ParserError, UnsetError
from .substitutions import SubstitutionContext, SubstitutionError

pp.ParserElement.enable_packrat()

_parser = None

# see https://stackoverflow.com/questions/43244861/pyparsing-infixnotation-optimization
# for a cleaner parser with functions


def _not_operator(value):
    return value is UNSET or isinstance(value, Unresolved) or not value


_UNARY_OPERATORS = {"+": lambda x: +x, "-": lambda x: -x, "~": lambda x: ~x, "not": _not_operator}

_BINARY_OPERATORS = {
    "**": lambda x, y: x**y,
    "*": lambda x, y: x * y,
    "/": lambda x, y: x / y,
    "//": lambda x, y: x // y,
    "+": lambda x, y: x + y,
    "-": lambda x, y: x - y,
    "<<": lambda x, y: x << y,
    ">>": lambda x, y: x >> y,
    "&": lambda x, y: x & y,
    "^": lambda x, y: x ^ y,
    "|": lambda x, y: x | y,
    "==": lambda x, y: x == y,
    "!=": lambda x, y: x != y,
    "<=": lambda x, y: x <= y,
    "<": lambda x, y: x < y,
    ">=": lambda x, y: x >= y,
    ">": lambda x, y: x > y,
    # 'is':   lambda x,y: x is y,
    # 'is not':   lambda x,y: x is not y,
    "in": lambda x, y: x in y,
    "not in": lambda x, y: x not in y,
    "and": lambda x, y: x and y,
    "or": lambda x, y: x or y,
}


class ResultsHandler(object):
    def evaluate(evaluator):
        pass

    def dump(self):
        pass

    @classmethod
    def pa(cls, s, l, t):
        return cls(*t[0])


class UnaryHandler(ResultsHandler):
    def __init__(self, op, arg):
        self.op, self.arg = op, arg
        assert op in _UNARY_OPERATORS
        self._op = _UNARY_OPERATORS[op]
        # not allows UNSET values
        self.allow_unset = op == "not"

    def evaluate(self, evaluator):
        arg = evaluator._evaluate_result(self.arg, allow_unset=self.allow_unset)
        if isinstance(arg, Unresolved):
            return arg
        return self._op(arg)

    def dump(self):
        return f"UnaryHandler({self.op}\n   {self.arg})"


class BinaryHandler(ResultsHandler):
    def __init__(self, arg1, op, arg2):
        self.op, self.arg1, self.arg2 = op, arg1, arg2
        assert op in _BINARY_OPERATORS
        self._op = _BINARY_OPERATORS[op]
        self.allow_unset = False

    def evaluate(self, evaluator):
        arg1, arg2 = (
            evaluator._evaluate_result(self.arg1, allow_unset=self.allow_unset),
            evaluator._evaluate_result(self.arg2, allow_unset=self.allow_unset),
        )
        if isinstance(arg1, Unresolved):
            return arg1
        if isinstance(arg2, Unresolved):
            return arg2
        return self._op(arg1, arg2)

    def dump(self):
        return f"BinaryHandler({self.arg1},\n   {self.op},\n{self.arg2})"

    @staticmethod
    def pa(s, l, t):
        # https://stackoverflow.com/questions/4571441/recursive-expressions-with-pyparsing
        initlen, incr = 3, 2
        t = t[0]
        ret = BinaryHandler(*t[:initlen])
        i = initlen
        while i < len(t):
            ret = BinaryHandler(ret, *t[i : i + incr])
            i += incr
        return ret


class GetItemHandler(ResultsHandler):
    def __init__(self, base):
        self.base, self.index = base[0], base[1]

    @staticmethod
    def pa(s, l, t):
        # https://stackoverflow.com/questions/4571441/recursive-expressions-with-pyparsing
        return GetItemHandler(*t)

    def evaluate(self, evaluator):
        base = evaluator._evaluate_result(self.base)
        index = evaluator._evaluate_result(self.index)
        if isinstance(base, Unresolved):
            return base
        if isinstance(index, Unresolved):
            return index
        return base[index]


class FunctionHandler(ResultsHandler):
    def __init__(self, func, *args):
        self.func, self.args = func, args
        self._func = getattr(self, func)
        if self._func is None:
            raise FormulaError(f"unknown function {func}")

    @staticmethod
    def pa(s, l, t):
        # https://stackoverflow.com/questions/4571441/recursive-expressions-with-pyparsing
        return FunctionHandler(*t[0])

    def evaluate(self, evaluator):
        return self._func(evaluator, self.args)

    def evaluate_generic_callable(self, evaluator, name, callable, args, min_args=None, max_args=None):
        if min_args is not None and len(args) < min_args:
            raise FormulaError(f"{'.'.join(evaluator.location)}: {name}() expects at least {min_args} argument(s)")
        if max_args is not None and len(args) > max_args:
            raise FormulaError(f"{'.'.join(evaluator.location)}: {name}() expects at most {max_args} argument(s)")
        eval_args = [evaluator._evaluate_result(arg) for arg in args]
        # if any argument is UNSET, return it as our result
        unsets = [arg for arg in eval_args if isinstance(arg, Unresolved)]
        if unsets:
            return unsets[0]
        return callable(*eval_args)

    def ERROR(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: ERROR() expects one argument, got {len(args)}")
        cond = evaluator._evaluate_result(args[0], allow_unset=True)
        raise FormulaError(f"ERROR: {cond}")

    def LIST(self, evaluator, args):
        def make_list(*x):
            return list(x)

        return self.evaluate_generic_callable(evaluator, "LIST", make_list, args)

    def LEN(self, evaluator, args):
        def make_len(x):
            return len(x)

        return self.evaluate_generic_callable(evaluator, "LEN", make_len, args, min_args=1, max_args=1)

    def RANGE(self, evaluator, args):
        def make_range(*x):
            return list(range(*x))

        return self.evaluate_generic_callable(evaluator, "RANGE", make_range, args, min_args=1, max_args=3)

    def VALID(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: VALID() expects one argument, got {len(args)}")
        try:
            result = evaluator._evaluate_result(args[0], allow_unset=True)
        except (TypeError, ValueError, IndexError):
            return False
        if isinstance(result, Unresolved):
            return False
        return result

    def TRY(self, evaluator, args):
        for arg in args:
            result = evaluator._evaluate_result(arg, allow_unset=True)
            if not isinstance(result, Unresolved):
                return result
        return result

    def MIN(self, evaluator, args):
        return self.evaluate_generic_callable(evaluator, "MIN", min, args, min_args=1)

    def MAX(self, evaluator, args):
        return self.evaluate_generic_callable(evaluator, "MAX", max, args, min_args=1)

    def GETITEM(self, evaluator, args):
        def get_item(x, y):
            return x[y]

        return self.evaluate_generic_callable(evaluator, "GETITEM", get_item, args, min_args=2, max_args=2)

    def IS_STR(self, evaluator, args):
        def is_str(x):
            return type(x) is str

        return self.evaluate_generic_callable(evaluator, "IS_STR", is_str, args, min_args=1, max_args=1)

    def IS_NUM(self, evaluator, args):
        def is_num(x):
            return isinstance(x, (bool, int, float, complex))

        return self.evaluate_generic_callable(evaluator, "IS_NUM", is_num, args, min_args=1, max_args=1)

    def CASES(self, evaluator, args):
        # set default case
        if len(args) % 2:
            default_case = args[-1]
            args = args[:-1]
        else:
            default_case = None
        # return first True case
        for i in range(0, len(args), 2):
            conditional, result = args[i : i + 2]
            cond = evaluator._evaluate_result(conditional, allow_unset=False)
            if cond:
                return evaluator._evaluate_result(result, allow_unset=True)
        # return default
        if default_case is None:
            return UNSET("no match in CASES()")
        return evaluator._evaluate_result(default_case, allow_unset=True)

    def IF(self, evaluator, args):
        if len(args) < 3 or len(args) > 4:
            raise FormulaError(f"{'.'.join(evaluator.location)}: IF() expects 3 or 4 arguments, got {len(args)}")
        conditional, if_true, if_false = args[:3]
        if_unset = args[3] if len(args) == 4 else UNSET("")

        cond = evaluator._evaluate_result(conditional, allow_unset=if_unset is not None)
        if isinstance(cond, Unresolved):
            if if_unset is None:
                raise SubstitutionError(f"{'.'.join(evaluator.location)}: '{cond.name}' is not defined")
            result = if_unset
        else:
            result = if_true if cond else if_false

        return evaluator._evaluate_result(result)

    def IFSET(self, evaluator, args):
        if len(args) < 1 or len(args) > 3:
            raise FormulaError(f"{'.'.join(evaluator.location)}: IFSET() expects 1 to 3 arguments, got {len(args)}")

        lookup, if_set, if_unset = list(args) + [None] * (3 - len(args))

        value = evaluator._evaluate_result(lookup, allow_unset=True)
        if isinstance(value, Unresolved):
            if is_missing(if_unset):
                return UNSET
            else:
                return evaluator._evaluate_result(if_unset)
        elif is_missing(if_set) or if_set == "SELF":
            return value
        else:
            return evaluator._evaluate_result(if_set)

    def GLOB(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: GLOB() expects 1 argument, got {len(args)}")
        pattern = evaluator._evaluate_result(args[0])
        if isinstance(pattern, Unresolved):
            return pattern
        return sorted(glob.glob(pattern))

    def EXISTS(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: EXISTS() expects 1 argument, got {len(args)}")
        pattern = evaluator._evaluate_result(args[0])
        if isinstance(pattern, Unresolved):
            return pattern
        return bool(glob.glob(pattern))

    def DIRNAME(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: DIRNAME() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if isinstance(path, Unresolved):
            return path
        if not isinstance(path, str):
            raise FormulaError(f"DIRNAME() expects a string, got a {str(type(path))}")
        return os.path.dirname(path)

    def BASENAME(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: BASENAME() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if isinstance(path, Unresolved):
            return path
        if not isinstance(path, str):
            raise FormulaError(f"BASENAME() expects a string, got a {str(type(path))}")
        return os.path.basename(path)

    def EXTENSION(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: EXTENSION() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if isinstance(path, Unresolved):
            return path
        if not isinstance(path, str):
            raise FormulaError(f"EXTENSION() expects a string, got a {str(type(path))}")
        return os.path.splitext(path)[1]

    def STRIPEXT(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: STRIPEXT() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if isinstance(path, Unresolved):
            return path
        if not isinstance(path, str):
            raise FormulaError(f"STRIPEXT() expects a string, got a {str(type(path))}")
        return os.path.splitext(path)[0]

    def NOSUBST(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: NOSUBST() expects 1 argument, got {len(args)}")
        return evaluator._evaluate_result(args[0], subst=False)

    def SORT(self, evaluator, args):
        return self._sort_impl(evaluator, args, "SORT")

    def RSORT(self, evaluator, args):
        return self._sort_impl(evaluator, args, "RSORT", reverse=True)

    def _sort_impl(self, evaluator, args, funcname, reverse=False):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: {funcname}() expects 1 argument, got {len(args)}")
        sortlist = evaluator._evaluate_result(args[0])
        if isinstance(sortlist, Unresolved):
            return sortlist
        if not isinstance(sortlist, (list, tuple)):
            raise FormulaError(f"{funcname}() expects a list, got a {str(type(sortlist))}")
        return sorted(sortlist, reverse=reverse)


def construct_parser():
    lparen = pp.Literal("(").suppress()
    rparen = pp.Literal(")").suppress()
    lbrack = pp.Literal("[").suppress()
    rbrack = pp.Literal("]").suppress()
    comma = pp.Literal(",").suppress()  # noqa: F841 - unusued
    period = pp.Literal(".").suppress()
    string = (pp.QuotedString('"') | pp.QuotedString("'"))("constant")
    UNSET = pp.Keyword("UNSET")("unset")
    SELF = pp.Keyword("SELF")("self_value")
    EMPTY = pp.Keyword("EMPTY")("empty")
    bool_false = (pp.Keyword("False") | pp.Keyword("false"))("bool_false").set_parse_action(lambda: [False])
    bool_true = (pp.Keyword("True") | pp.Keyword("true"))("bool_true").set_parse_action(lambda: [True])

    boolean = (bool_true | bool_false)("constant")
    number = pp.common.number("constant")

    fieldname = pp.Word(pp.alphas + "_", pp.alphanums + "_-@*?")
    nested_field = pp.Group(fieldname + pp.OneOrMore(period + fieldname))("namespace_lookup")
    anyseq = pp.CharsNotIn(",)")("constant")

    # allow expression to be used recursively
    expr = pp.Forward()

    # functions -- get all all-uppercase members from FunctionHandler
    anyseq_funcnames = {"GLOB", "EXISTS", "ERROR"}
    all_funcnames = set(
        func for func in dir(FunctionHandler) if callable(getattr(FunctionHandler, func)) and func.upper() == func
    )
    all_funcnames -= anyseq_funcnames

    functions = reduce(operator.or_, map(pp.Keyword, all_funcnames))
    # these functions take one argument, which could also be a sequence
    anyseq_functions = reduce(operator.or_, map(pp.Keyword, anyseq_funcnames))

    atomic_value = boolean | UNSET | EMPTY | nested_field | string | number

    function_call_anyseq = pp.Group(anyseq_functions + lparen + (expr | anyseq) + rparen).set_parse_action(
        FunctionHandler.pa
    )
    function_call = pp.Group(functions + lparen + pp.Opt(pp.DelimitedList(expr | SELF)) + rparen).set_parse_action(
        FunctionHandler.pa
    )

    operators = [
        ((lbrack + expr + rbrack), 1, pp.opAssoc.LEFT, GetItemHandler.pa),
        (pp.Literal("**"), 2, pp.opAssoc.LEFT, BinaryHandler.pa),
        (pp.Literal("-") | pp.Literal("+") | pp.Literal("~"), 1, pp.opAssoc.RIGHT, UnaryHandler.pa),
        (pp.Literal("*") | pp.Literal("//") | pp.Literal("/") | pp.Literal("%"), 2, pp.opAssoc.LEFT, BinaryHandler.pa),
        (pp.Literal("+") | pp.Literal("-"), 2, pp.opAssoc.LEFT, BinaryHandler.pa),
        (pp.Literal("<<") | pp.Literal(">>"), 2, pp.opAssoc.LEFT, BinaryHandler.pa),
        (pp.Literal("&"), 2, pp.opAssoc.LEFT, BinaryHandler.pa),
        (pp.Literal("^"), 2, pp.opAssoc.LEFT, BinaryHandler.pa),
        (pp.Literal("|"), 2, pp.opAssoc.LEFT, BinaryHandler.pa),
        (
            reduce(operator.or_, map(pp.Literal, ("==", "!=", ">=", "<=", ">", "<"))),
            2,
            pp.opAssoc.LEFT,
            BinaryHandler.pa,
        ),
        (pp.CaselessKeyword("in") | pp.CaselessKeyword("not in"), 2, pp.opAssoc.LEFT, BinaryHandler.pa),
        (pp.CaselessKeyword("not"), 1, pp.opAssoc.RIGHT, UnaryHandler.pa),
        (pp.CaselessKeyword("and") | pp.CaselessKeyword("or"), 2, pp.opAssoc.LEFT, BinaryHandler.pa),
    ]

    infix = pp.infix_notation(atomic_value | function_call | function_call_anyseq | nested_field, operators)(
        "subexpression"
    )

    expr <<= infix

    return expr


if _parser is None:
    _parser = construct_parser()

_parse_cache = {}


def parse_string(text: str, location: List[str] = []):
    """parses a formula, returns ParseResults object. Implements cache.

    Args:
        text (str): formula to parse
        location (List[str], optional): Hierarchical location, used for erro messages

    Raises:
        ParserError: on any parse error

    Returns:
        ParseResults: on successful parse
    """
    parse_results = _parse_cache.get(text)
    if parse_results is None:
        try:
            parse_results = _parser.parse_string(text, parse_all=True)
        except Exception as exc:
            parse_results = ParserError(f"{'.'.join(location)}: error parsing formula ({exc})")
        _parse_cache[text] = parse_results

    if isinstance(parse_results, Exception):
        raise parse_results

    return parse_results


def is_missing(result):
    return result is None


class SELF(object):
    pass


class Evaluator(object):
    def __init__(
        self,
        ns: Dict[str, Any],
        subst_context: typing.Optional[SubstitutionContext] = None,
        allow_unresolved: bool = False,
        location: List[str] = [],
    ):
        self.ns = ns
        self.subst_context = subst_context
        self.location = location
        self.allow_unresolved = allow_unresolved

    def _resolve(self, value, in_formula=True, subst=True):
        if type(value) is str:
            if in_formula and value == "SELF":
                return SELF
            elif in_formula and value == "UNSET":
                return UNSET
            elif self.subst_context is not None and subst:
                try:
                    value = self.subst_context.evaluate(value, location=self.location)
                except (KeyError, AttributeError) as exc:
                    raise SubstitutionError(f"{value}: invalid key {exc}")
                except Exception as exc:
                    raise SubstitutionError(f"{value}: {exc}")
        return value

    def empty(self, *args, **kw):
        return ""

    def unset(self, *args, **kw):
        return UNSET

    def self_value(self, *args, **kw):
        return SELF

    def constant(self, value, subst=True, **kw):
        return self._resolve(value, subst=subst)

    def subexpression(self, value, subst=True):
        return self._evaluate_result(value, allow_unset=True, subst=subst)

    def namespace_lookup(self, *args, subst=True):
        if len(args) == 1 and type(args[0]) is pp.ParseResults:
            args = args[0]
            assert args._name == "namespace_lookup"
        value = self.ns
        fields = list(args)
        while fields:
            fld = fields.pop(0)
            # check for wildcards
            if fld not in value and ("*" in fld or "?" in fld):
                names = sorted(fnmatch.filter(value.keys(), fld))
                if names:
                    fld = names[-1]
            # last element allowed to be UNSET, otherwise substitution error
            if fld not in value:
                if fields:
                    raise SubstitutionError(f"{'.'.join(self.location)}: '{fld}' undefined (in '{'.'.join(args)}')")
                else:
                    return UNSET(".".join(args))
            # this can still throw an error if a nested interpolation is invoked
            try:
                value = value.get(fld, subst=subst)
            except (KeyError, AttributeError) as exc:
                raise SubstitutionError(f"{'.'.join(self.location)}: '{'.'.join(args)}' unresolved (at '{exc}')")
        return self._resolve(value, subst=subst)

    def _evaluate_result(self, parse_result, allow_unset=False, subst=True):
        allow_unset = allow_unset or self.allow_unresolved
        # if result is a handler, use evaluate
        if isinstance(parse_result, ResultsHandler):
            value = parse_result.evaluate(self)

        # if result is already a constant, resolve it
        elif type(parse_result) is not pp.ParseResults:
            return self._resolve(parse_result, subst=subst)

        # lookup processing method based on name
        else:
            method = parse_result.getName()
            assert method is not None
            if not hasattr(self, method):
                raise ParserError(
                    f"{'.'.join(self.location)}: don't know how to deal with an element of type '{method}'"
                )

            try:
                value = getattr(self, method)(*parse_result, subst=subst)
            except SubstitutionError as exc:
                if allow_unset:
                    return UNSET("", [exc])
                raise

        if isinstance(value, Unresolved) and not allow_unset:
            raise UnsetError(f"'{value.value}' undefined", value.errors)

        return value

    def evaluate(self, value: Any, sublocation: List[str] = []) -> Any:
        """evaluates a single value, which can be a string with a formula and/or substitutions

        Args:
            value (Any): value to be evaluated. If not a str, returned as is
            sublocation (List[str]): location of value inside of nested container, as a list
                    of strings. Defaults to []

        Raises:
            ParserError: error parsing formula
            FormulaError: error evaluating formula
            SubstitutionError: unresolved symbol in formula or {}-substitution

        Returns:
            Any: result of evaluation
        """

        if type(value) is not str:
            return value

        loclen = len(self.location)
        self.location += sublocation

        try:
            if value.startswith("="):
                if value.startswith("=="):
                    return self._resolve(value[1:], in_formula=False)
                else:
                    try:
                        parse_results = parse_string(value[1:])
                    except Exception as exc:
                        raise ParserError(f"{'.'.join(self.location)}: error parsing formula '{value}'", exc)

                    try:
                        return self._evaluate_result(parse_results, allow_unset=True)
                    except Exception as exc:
                        raise FormulaError(f"{'.'.join(self.location)}: evaluation of '{value}' failed", exc, tb=True)
            else:
                try:
                    return self._resolve(value, in_formula=False)
                except Exception as exc:
                    raise SubstitutionError(f"{'.'.join(self.location)}: evaluation of '{value}' failed", exc)
        finally:
            self.location = self.location[:loclen]

    def evaluate_object(
        self,
        obj: Any,
        sublocation: List[str] = [],
        raise_substitution_errors: bool = True,
        recursion_level: int = 1,
        verbose: bool = False,
    ) -> Any:
        """evaluates object, which can be a nested container, in which case string elements
        are evaluated recursively. Evaluations are done in place.

        Args:
            obj (Any): object to evaluate
            sublocation (List[str]): location of value inside of nested container, as a list
                    of strings. Defaults to []
            raise_substitution_errors (bool, optional): raise substitution errors, instead of returning
                Unresolved values. Defaults to True.
            recursion_level (int, optional): Limits recursion level into subcontainers. Defaults to 1,
                which means recurse once.
            verbose (bool, optional): Prints debug messages. Defaults to False.

        Raises:
            SubstitutionError: error in substitutions

        Returns:
            Any: object with evaluations
        """
        # string? evaluate directly and return
        if type(obj) is str:
            try:
                value = self.evaluate(obj, sublocation=sublocation)
                if isinstance(value, Unresolved) and (raise_substitution_errors or not self.allow_unresolved):
                    raise SubstitutionError(f"{'.'.join(sublocation)}: unresolved substitution", value.errors)
                return value
            except AttributeError as err:
                if raise_substitution_errors or not self.allow_unresolved:
                    raise SubstitutionError(f"{'.'.join(sublocation)}: substitution error", [err])
                return Unresolved(errors=[err])
            except SubstitutionError as err:
                if raise_substitution_errors or not self.allow_unresolved:
                    raise SubstitutionError(f"{'.'.join(sublocation)}: substitution error", [err])
                return Unresolved(errors=[err])

        # helper function
        def update(value, sloc):
            if isinstance(value, Unresolved):
                return value, False
            subloc = sublocation + [sloc]
            if verbose:
                print(f"{subloc}: {value} ...")
            new_value = self.evaluate_object(
                value,
                raise_substitution_errors=raise_substitution_errors,
                recursion_level=recursion_level,
                verbose=verbose,
                sublocation=subloc,
            )
            if verbose:
                print(f"{subloc}: {value} -> {new_value}")
            # UNSET return means delete or revert to default
            if new_value is UNSET:
                raise SubstitutionError(f"{'.'.join(self.location + subloc)}: UNSET not allowed here")
            # compare
            if isinstance(value, (dict, DictConfig, list, ListConfig)) or dataclasses.is_dataclass(value):
                updated = value is not new_value
            else:
                updated = value != new_value
            return new_value, updated

        obj_out = obj
        # recurse into containers?
        if recursion_level:
            recursion_level -= 1
            # use evaluate_dict() to recurse into dicts
            if isinstance(obj, (dict, DictConfig)):
                for key, value in obj.items():
                    new_value, value_updated = update(value, key)
                    new_key = self.evaluate(key, sublocation=sublocation)
                    if new_key != key:
                        value_updated = True
                    if value_updated:
                        if obj_out is obj:
                            obj_out = obj.copy()
                        if new_key != key:
                            del obj_out[key]
                            key = new_key
                        obj_out[key] = new_value
            # recurse into lists
            elif isinstance(obj, (list, ListConfig)):
                for i, value in enumerate(obj):
                    new_value, updated = update(value, f"#{i}")
                    if updated:
                        if obj_out is obj:
                            obj_out = obj.copy()
                        obj_out[i] = new_value
            # recurse into dataclasses
            elif dataclasses.is_dataclass(obj):
                newvals = {}
                for fld in dataclasses.fields(obj):
                    value = getattr(obj, fld.name)
                    new_value, updated = update(value, fld.name)
                    if updated:
                        newvals[fld.name] = new_value
                if newvals:
                    obj_out = dataclasses.replace(obj, **newvals)

        return obj_out

    def evaluate_dict(
        self,
        params: Dict[str, Any],
        corresponding_ns: typing.Optional[Dict[str, Any]] = None,
        defaults: Dict[str, Any] = {},
        sublocation=[],
        raise_substitution_errors: bool = True,
        collapse_substitution_errors: bool = False,  # true for subcontainers
        subcontainer_type: typing.Optional[str] = None,
        recursive: bool = True,
        verbose: bool = False,
    ) -> Dict[str, Any]:
        """evaluates dict of parameters, which can contain nested containers which can be evaluated recursively.
        Returns new dict.

        Args:
            params (Dict[str, Any]): parameters to evaluate
            corresponding_ns (Optional[Dict[str, Any]], optional): corresponding namespace, into which evluations
                are propagated. Defaults to None.
            defaults (Dict[str, Any], optional): dictionary of defaults; UNSET values are cause defaults to
                be substituted in.
            sublocation (list, optional): location inside of nested container, as a list
                of strings. Defaults to []
            raise_substitution_errors (bool, optional): raises substitution errors. Defaults to True.
                if False, substitution errors are assigned to dict as Unresolved values.
            collapse_substitution_errors (bool, optional): if True, a substituion error inside the dict
                causes an Unresolved value to be returned instead of the dict. Defaults to False.
            subcontainer_type (str): type of subcontainer being evaluated. Defaults to None. Must be set
                if collapse_substitution_errors is True.
            recursive (bool, optional): recurse into subcontainers. Defaults to True.
            verbose (bool, optional): print debug messages. Defaults to False.

        Returns:
            Dict[str, Any]: _descopy of input dict with substitutions performed
        """
        if collapse_substitution_errors:
            assert subcontainer_type is not None
        params_out = params
        for name, value in list(params.items()):
            if isinstance(value, Unresolved):
                continue
            #
            retry = True
            while retry:
                retry = False
                if verbose:  # or type(value) is UNSET:
                    print(f"{name}: {value} ...")
                if type(value) is str:
                    try:
                        new_value = self.evaluate(value, sublocation=sublocation + [name])
                    except (AttributeError, SubstitutionError, ParserError, FormulaError) as err:
                        if raise_substitution_errors:
                            raise
                        new_value = Unresolved(errors=[str(err)])
                    if verbose:
                        print(f"{name}: {value} -> {new_value}")
                    # UNSET return means delete or revert to default
                    if new_value is UNSET:
                        if subcontainer_type:
                            raise SubstitutionError(f"{'.'.join(self.location + sublocation)}: UNSET not allowed here")
                        if params_out is params:
                            params_out = params.copy()
                        # if value is in defaults and is different, try to evaluate that instead
                        if name in defaults and defaults[name] is not UNSET and defaults[name] != value:
                            value = params_out[name] = defaults[name]
                            if corresponding_ns:
                                corresponding_ns[name] = defaults[name]
                            retry = True
                        else:
                            if name in params_out:
                                del params_out[name]
                            if corresponding_ns and name in corresponding_ns:
                                del corresponding_ns[name]
                    elif new_value is not value and new_value != value:
                        if params_out is params:
                            params_out = params.copy()
                        params_out[name] = new_value
                        if corresponding_ns:
                            corresponding_ns[name] = new_value
                elif isinstance(value, (dict, DictConfig)) and recursive:
                    value = self.evaluate_dict(
                        value,
                        corresponding_ns,
                        defaults,
                        sublocation=sublocation + [name],
                        raise_substitution_errors=raise_substitution_errors,
                        collapse_substitution_errors=True,
                        subcontainer_type="Dict",
                        recursive=True,
                        verbose=verbose,
                    )
                    params_out[name] = value
                    if corresponding_ns:
                        corresponding_ns[name] = value
                elif isinstance(value, (list, ListConfig)) and recursive:
                    # convert list to dict, and evaluate
                    proxy_dict = self.evaluate_dict(
                        {f"[{i}]": v for i, v in enumerate(value)},
                        corresponding_ns,
                        defaults,
                        sublocation=sublocation + [name],
                        raise_substitution_errors=raise_substitution_errors,
                        collapse_substitution_errors=True,
                        subcontainer_type="List",
                        recursive=True,
                        verbose=verbose,
                    )
                    if isinstance(proxy_dict, Unresolved):
                        value = proxy_dict
                    else:
                        value = type(value)(list(proxy_dict.values()))
                    params_out[name] = value
                    if corresponding_ns:
                        corresponding_ns[name] = value

        if collapse_substitution_errors:
            errors = []
            for elem in params_out.values():
                if isinstance(elem, Unresolved):
                    errors += elem.errors
            if errors:
                return Unresolved(f"unresolved {subcontainer_type} elements", errors)

        return params_out


if __name__ == "__main__":
    pass
